"""Phonopy input and command option tools."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import warnings

import numpy as np


def fracval(frac):
    """Return floating point value from rational."""
    if frac.find("/") == -1:
        return float(frac)
    else:
        x = frac.split("/")
        return float(x[0]) / float(x[1])


class Settings:
    """Phonopy settings container.

    This works almost like a dictionary.
    Method names without 'set_' and 'get_' and keys of self._v have to be same.

    """

    _default = {
        "band_indices": None,
        "band_paths": None,
        "band_points": None,
        "cell_filename": None,
        "chemical_symbols": None,
        "classical": False,
        "cutoff_frequency": None,
        "displacement_distance": None,
        "displacement_distance_max": None,
        "dm_decimals": None,
        "calculator": None,
        "create_displacements": False,
        "fc_calculator": None,
        "fc_calculator_options": None,
        "fc_decimals": None,
        "fc_symmetry": False,
        "frequency_pitch": None,
        "frequency_conversion_factor": None,
        "frequency_scale_factor": None,
        "group_velocity_delta_q": None,
        "hdf5_compression": "gzip",
        "is_band_const_interval": False,
        "is_diagonal_displacement": True,
        "is_eigenvectors": False,
        "is_mesh_symmetry": True,
        "is_nac": False,
        "is_plusminus_displacement": "auto",
        "is_symmetry": True,
        "is_tetrahedron_method": True,
        "is_time_reversal_symmetry": True,
        "is_trigonal_displacement": False,
        "magnetic_moments": None,
        "masses": None,
        "mesh_numbers": None,
        "mlp_params": None,
        "nac_method": None,
        "nac_q_direction": None,
        "num_frequency_points": None,
        "primitive_matrix": None,
        "qpoints": None,
        "random_displacements": None,
        "random_seed": None,
        "read_qpoints": False,
        "save_params": False,
        "sigma": None,
        "supercell_matrix": None,
        "symmetry_tolerance": None,
        "max_temperature": 1000,
        "min_temperature": 0,
        "temperature_step": 10,
        "use_pypolymlp": False,
    }

    def __init__(self, default=None):
        """Init method.

        Note
        ----
        self.default is used in inherited classes.

        """
        self.default = Settings._default.copy()
        if default is not None:
            self.default.update(default)
        self._v = self.default.copy()

    def __getattr__(self, attr):
        """Return dict value of key attr."""
        return self._v[attr]

    def set_band_paths(self, val):
        """Set band_paths."""
        self._v["band_paths"] = val

    def set_band_points(self, val):
        """Set band_points."""
        self._v["band_points"] = val

    def set_band_indices(self, val):
        """Set band_indices."""
        self._v["band_indices"] = val

    def set_cell_filename(self, val):
        """Set cell_filename."""
        self._v["cell_filename"] = val

    def set_chemical_symbols(self, val):
        """Set chemical_symbols."""
        self._v["chemical_symbols"] = val

    def set_classical(self, val):
        """Set classical."""
        self._v["classical"] = val

    def set_create_displacements(self, val):
        """Set create_displacements."""
        self._v["create_displacements"] = val

    def set_cutoff_frequency(self, val):
        """Set cutoff_frequency."""
        self._v["cutoff_frequency"] = val

    def set_dm_decimals(self, val):
        """Set dm_decimals."""
        self._v["dm_decimals"] = val

    def set_displacement_distance(self, val):
        """Set displacement_distance."""
        self._v["displacement_distance"] = val

    def set_displacement_distance_max(self, val):
        """Set displacement_distance_max."""
        self._v["displacement_distance_max"] = val

    def set_calculator(self, val):
        """Set calculator."""
        self._v["calculator"] = val

    def set_fc_calculator(self, val):
        """Set fc_calculator."""
        self._v["fc_calculator"] = val

    def set_fc_calculator_options(self, val):
        """Set fc_calculator_options."""
        self._v["fc_calculator_options"] = val

    def set_fc_symmetry(self, val):
        """Set fc_symmetry."""
        self._v["fc_symmetry"] = val

    def set_fc_decimals(self, val):
        """Set fc_decimals."""
        self._v["fc_decimals"] = val

    def set_frequency_conversion_factor(self, val):
        """Set frequency_conversion_factor."""
        self._v["frequency_conversion_factor"] = val

    def set_frequency_pitch(self, val):
        """Set frequency_pitch."""
        self._v["frequency_pitch"] = val

    def set_frequency_scale_factor(self, val):
        """Set frequency_scale_factor."""
        self._v["frequency_scale_factor"] = val

    def set_group_velocity_delta_q(self, val):
        """Set group_velocity_delta_q."""
        self._v["group_velocity_delta_q"] = val

    def set_hdf5_compression(self, val):
        """Set hdf5_compression."""
        self._v["hdf5_compression"] = val

    def set_is_band_const_interval(self, val):
        """Set is_band_const_interval."""
        self._v["is_band_const_interval"] = val

    def set_is_diagonal_displacement(self, val):
        """Set is_diagonal_displacement."""
        self._v["is_diagonal_displacement"] = val

    def set_is_eigenvectors(self, val):
        """Set is_eigenvectors."""
        self._v["is_eigenvectors"] = val

    def set_is_mesh_symmetry(self, val):
        """Set is_mesh_symmetry."""
        self._v["is_mesh_symmetry"] = val

    def set_is_nac(self, val):
        """Set is_nac."""
        self._v["is_nac"] = val

    def set_is_plusminus_displacement(self, val):
        """Set is_plusminus_displacement."""
        self._v["is_plusminus_displacement"] = val

    def set_is_tetrahedron_method(self, val):
        """Set is_tetrahedron_method."""
        self._v["is_tetrahedron_method"] = val

    def set_is_trigonal_displacement(self, val):
        """Set is_trigonal_displacement."""
        self._v["is_trigonal_displacement"] = val

    def set_is_symmetry(self, val):
        """Set is_symmetry."""
        self._v["is_symmetry"] = val

    def set_magnetic_moments(self, val):
        """Set magnetic_moments."""
        self._v["magnetic_moments"] = val

    def set_masses(self, val):
        """Set masses."""
        self._v["masses"] = val

    def set_max_temperature(self, val):
        """Set max_temperature."""
        self._v["max_temperature"] = val

    def set_mesh_numbers(self, val):
        """Set mesh_numbers."""
        self._v["mesh_numbers"] = val

    def set_min_temperature(self, val):
        """Set min_temperature."""
        self._v["min_temperature"] = val

    def set_mlp_params(self, val):
        """Set mlp_params."""
        self._v["mlp_params"] = val

    def set_nac_method(self, val):
        """Set nac_method."""
        self._v["nac_method"] = val

    def set_nac_q_direction(self, val):
        """Set nac_q_direction."""
        self._v["nac_q_direction"] = val

    def set_num_frequency_points(self, val):
        """Set num_frequency_points."""
        self._v["num_frequency_points"] = val

    def set_primitive_matrix(self, val):
        """Set primitive_matrix."""
        self._v["primitive_matrix"] = val

    def set_qpoints(self, val):
        """Set qpoints."""
        self._v["qpoints"] = val

    def set_random_displacements(self, val):
        """Set random_displacements."""
        self._v["random_displacements"] = val

    def set_random_seed(self, val):
        """Set random_seed."""
        self._v["random_seed"] = val

    def set_read_qpoints(self, val):
        """Set read_qpoints."""
        self._v["read_qpoints"] = val

    def set_save_params(self, val):
        """Set save_params."""
        self._v["save_params"] = val

    def set_sigma(self, val):
        """Set sigma."""
        self._v["sigma"] = val

    def set_supercell_matrix(self, val):
        """Set supercell_matrix."""
        self._v["supercell_matrix"] = val

    def set_symmetry_tolerance(self, val):
        """Set symmetry_tolerance."""
        self._v["symmetry_tolerance"] = val

    def set_temperature_step(self, val):
        """Set temperature_step."""
        self._v["temperature_step"] = val

    def set_is_time_reversal_symmetry(self, val):
        """Set is_time_reversal_symmetry."""
        self._v["is_time_reversal_symmetry"] = val

    def set_use_pypolymlp(self, val):
        """Set use_pypolymlp."""
        self._v["use_pypolymlp"] = val


# Parse phonopy setting filen
class ConfParser:
    """Phonopy conf file parser.

    This class became complicated due to historical reason. Data flow is as follows:

        self._confs -> self._parameters -> self._settings.

    Configurations are obtained either from conf file or command options.
    These are initially merged to ``self._confs``. Then the data are
    checked and stored in ``self._parameters`` in appropriate format.
    Finally they are passed to ``self._settings``.

    """

    def __init__(self, filename=None, args=None):
        """Init method."""
        self._confs = {}
        self._parameters = {}
        self._args = args
        self._filename = filename
        if "_settings" not in self.__dict__:
            self._settings = None

    @property
    def confs(self):
        """Return configuration dict."""
        return self._confs

    def get_configures(self):
        """Return configuration dict."""
        warnings.warn(
            "ConfParser.get_configures() is deprecated.Use confs attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.confs

    @property
    def settings(self):
        """Return Settings class instance."""
        return self._settings

    def get_settings(self):
        """Return Settings class instance."""
        warnings.warn(
            "ConfParser.get_settings() is deprecated.Use settings attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.settings

    def setting_error(self, message):
        """Show error message."""
        print(message)
        print("Please check the setting tags and options.")
        sys.exit(1)

    def read_file(self):
        """Read conf file."""
        with open(self._filename, "r") as file:
            is_continue = False
            left = None

            for line in file:
                if line.strip() == "":
                    is_continue = False
                    continue

                if line.strip()[0] == "#":
                    is_continue = False
                    continue

                if is_continue and left is not None:
                    self._confs[left] += line.strip()
                    self._confs[left] = self._confs[left].replace("+++", " ")
                    is_continue = False

                if line.find("=") != -1:
                    left, right = [x.strip() for x in line.split("=")]
                    self._confs[left.lower()] = right

                if line.find("+++") != -1:
                    is_continue = True

    def read_options(self):
        """Read options from ArgumentParser class instance.

        This is the interface layer to make settings from command options to be
        consistent with settings from conf file.

        """
        arg_list = vars(self._args)
        if "band_indices" in arg_list:
            band_indices = self._args.band_indices
            if band_indices is not None:
                if isinstance(band_indices, list):
                    self._confs["band_indices"] = " ".join(band_indices)
                else:
                    self._confs["band_indices"] = band_indices

        if "band_paths" in arg_list:
            if self._args.band_paths is not None:
                if isinstance(self._args.band_paths, list):
                    self._confs["band"] = " ".join(self._args.band_paths)
                else:
                    self._confs["band"] = self._args.band_paths

        if "band_points" in arg_list:
            if self._args.band_points is not None:
                self._confs["band_points"] = self._args.band_points

        if "cell_filename" in arg_list:
            if self._args.cell_filename is not None:
                self._confs["cell_filename"] = self._args.cell_filename

        if "classical" in arg_list:
            if self._args.classical:
                self._confs["classical"] = ".true."

        if "cutoff_frequency" in arg_list:
            if self._args.cutoff_frequency:
                self._confs["cutoff_frequency"] = self._args.cutoff_frequency

        if "displacement_distance" in arg_list:
            if self._args.displacement_distance is not None:
                self._confs["displacement_distance"] = self._args.displacement_distance

        if "displacement_distance_max" in arg_list:
            if self._args.displacement_distance_max is not None:
                self._confs["displacement_distance_max"] = (
                    self._args.displacement_distance_max
                )

        if "dynamical_matrix_decimals" in arg_list:
            if self._args.dynamical_matrix_decimals:
                self._confs["dm_decimals"] = self._args.dynamical_matrix_decimals

        if "calculator" in arg_list:
            if self._args.calculator:
                self._confs["calculator"] = self._args.calculator

        if "fc_calculator" in arg_list:
            if self._args.fc_calculator:
                self._confs["fc_calculator"] = self._args.fc_calculator

        if "fc_calculator_options" in arg_list:
            fc_calc_opt = self._args.fc_calculator_options
            if fc_calc_opt:
                self._confs["fc_calculator_options"] = fc_calc_opt

        if "fc_symmetry" in arg_list:
            if self._settings.default["fc_symmetry"]:
                if self._args.fc_symmetry is False:
                    self._confs["fc_symmetry"] = ".false."
            else:
                if self._args.fc_symmetry:
                    self._confs["fc_symmetry"] = ".true."

        if "force_constants_decimals" in arg_list:
            if self._args.force_constants_decimals:
                self._confs["fc_decimals"] = self._args.force_constants_decimals

        if "fpitch" in arg_list:
            if self._args.fpitch:
                self._confs["fpitch"] = self._args.fpitch

        if "frequency_conversion_factor" in arg_list:
            freq_factor = self._args.frequency_conversion_factor
            if freq_factor:
                self._confs["frequency_conversion_factor"] = freq_factor

        if "frequency_scale_factor" in arg_list:
            freq_scale = self._args.frequency_scale_factor
            if freq_scale is not None:
                self._confs["frequency_scale_factor"] = freq_scale

        if "gv_delta_q" in arg_list:
            if self._args.gv_delta_q:
                self._confs["gv_delta_q"] = self._args.gv_delta_q

        if "hdf5_compression" in arg_list:
            if self._args.hdf5_compression:
                self._confs["hdf5_compression"] = self._args.hdf5_compression

        if "is_band_const_interval" in arg_list:
            if self._args.is_band_const_interval:
                self._confs["band_const_interval"] = ".true."

        if "is_displacement" in arg_list:
            if self._args.is_displacement:
                self._confs["create_displacements"] = ".true."

        if "is_eigenvectors" in arg_list:
            if self._args.is_eigenvectors:
                self._confs["eigenvectors"] = ".true."

        if "is_nac" in arg_list:
            if self._settings.default["is_nac"]:  # Check default settings
                if self._args.is_nac is False:
                    self._confs["nac"] = ".false."
            else:
                if self._args.is_nac:
                    self._confs["nac"] = ".true."

        if "is_nodiag" in arg_list:
            if self._args.is_nodiag:
                self._confs["diag"] = ".false."

        if "is_nomeshsym" in arg_list:
            if self._args.is_nomeshsym:
                self._confs["mesh_symmetry"] = ".false."

        if "is_nosym" in arg_list:
            if self._args.is_nosym:
                self._confs["symmetry"] = ".false."

        if "is_plusminus_displacements" in arg_list:
            if self._args.is_plusminus_displacements:
                self._confs["pm"] = ".true."

        if "is_trigonal_displacements" in arg_list:
            if self._args.is_trigonal_displacements:
                self._confs["trigonal"] = ".true."

        if "masses" in arg_list:
            if self._args.masses is not None:
                if isinstance(self._args.masses, list):
                    self._confs["mass"] = " ".join(self._args.masses)
                else:
                    self._confs["mass"] = self._args.masses

        if "magmoms" in arg_list:
            if self._args.magmoms is not None:
                if isinstance(self._args.magmoms, list):
                    self._confs["magmom"] = " ".join(self._args.magmoms)
                else:
                    self._confs["magmom"] = self._args.magmoms

        if "mesh_numbers" in arg_list:
            mesh = self._args.mesh_numbers
            if mesh is not None:
                if isinstance(mesh, list):
                    self._confs["mesh_numbers"] = " ".join(mesh)
                else:
                    self._confs["mesh_numbers"] = mesh

        if "mlp_params" in arg_list:
            mlp_params = self._args.mlp_params
            if mlp_params:
                self._confs["mlp_params"] = mlp_params

        if "num_frequency_points" in arg_list:
            opt_num_freqs = self._args.num_frequency_points
            if opt_num_freqs:
                self._confs["num_frequency_points"] = opt_num_freqs

        # For backword compatibility
        if "primitive_axis" in arg_list:
            if self._args.primitive_axis is not None:
                if isinstance(self._args.primitive_axis, list):
                    primitive_axes = " ".join(self._args.primitive_axis)
                    self._confs["primitive_axes"] = primitive_axes
                else:
                    self._confs["primitive_axes"] = self._args.primitive_axis

        if "primitive_axes" in arg_list:
            if self._args.primitive_axes:
                if isinstance(self._args.primitive_axes, list):
                    primitive_axes = " ".join(self._args.primitive_axes)
                    self._confs["primitive_axes"] = primitive_axes
                else:
                    self._confs["primitive_axes"] = self._args.primitive_axes

        if "qpoints" in arg_list:
            if self._args.qpoints is not None:
                if isinstance(self._args.qpoints, list):
                    self._confs["qpoints"] = " ".join(self._args.qpoints)
                else:
                    self._confs["qpoints"] = self._args.qpoints

        if "nac_q_direction" in arg_list:
            q_dir = self._args.nac_q_direction
            if q_dir is not None:
                if isinstance(q_dir, list):
                    self._confs["q_direction"] = " ".join(q_dir)
                else:
                    self._confs["q_direction"] = q_dir

        if "nac_method" in arg_list:
            if self._args.nac_method is not None:
                self._confs["nac_method"] = self._args.nac_method

        if "random_displacements" in arg_list:
            nrand = self._args.random_displacements
            if nrand:
                self._confs["random_displacements"] = nrand

        if "random_seed" in arg_list:
            if self._args.random_seed:
                seed = self._args.random_seed
                if np.issubdtype(type(seed), np.integer) and seed >= 0 and seed < 2**32:
                    self._confs["random_seed"] = seed

        if "read_qpoints" in arg_list:
            if self._args.read_qpoints:
                self._confs["read_qpoints"] = ".true."

        if "save_params" in arg_list:
            if self._args.save_params:
                self._confs["save_params"] = ".true."

        if "supercell_dimension" in arg_list:
            dim = self._args.supercell_dimension
            if dim is not None:
                if isinstance(dim, list):
                    self._confs["dim"] = " ".join(dim)
                else:
                    self._confs["dim"] = dim

        if "sigma" in arg_list:
            if self._args.sigma is not None:
                if isinstance(self._args.sigma, list):
                    self._confs["sigma"] = " ".join(self._args.sigma)
                else:
                    self._confs["sigma"] = self._args.sigma

        if "symmetry_tolerance" in arg_list:
            if self._args.symmetry_tolerance:
                symtol = self._args.symmetry_tolerance
                self._confs["symmetry_tolerance"] = symtol

        if "tmax" in arg_list:
            if self._args.tmax:
                self._confs["tmax"] = self._args.tmax

        if "tmin" in arg_list:
            if self._args.tmin:
                self._confs["tmin"] = self._args.tmin

        if "tstep" in arg_list:
            if self._args.tstep:
                self._confs["tstep"] = self._args.tstep

        from phonopy.interface.calculator import get_interface_mode

        calculator = get_interface_mode(arg_list)
        if calculator:
            self._confs["calculator"] = calculator

        if "use_alm" in arg_list:
            if self._args.use_alm:
                self._confs["fc_calculator"] = "alm"

        if "use_symfc" in arg_list:
            if self._args.use_symfc:
                self._confs["fc_calculator"] = "symfc"

        if "use_pypolymlp" in arg_list:
            if self._args.use_pypolymlp:
                self._confs["use_pypolymlp"] = ".true."

    def parse_conf(self):
        """Add treatments to settings from conf file or command options.

        The results are stored in ``self._parameters[key] = val``.

        """
        confs = self._confs

        for conf_key in confs.keys():
            if conf_key == "band_indices":
                vals = []
                for sum_set in confs["band_indices"].split(","):
                    vals.append([int(x) - 1 for x in sum_set.split()])
                self.set_parameter("band_indices", vals)

            if conf_key == "cell_filename":
                self.set_parameter("cell_filename", confs["cell_filename"])

            if conf_key == "create_displacements":
                if confs["create_displacements"].lower() == ".true.":
                    self.set_parameter("create_displacements", True)
                elif confs["create_displacements"].lower() == ".false.":
                    self.set_parameter("create_displacements", False)

            if conf_key == "dim":
                matrix = [int(x) for x in confs["dim"].split()]
                if len(matrix) == 9:
                    matrix = np.array(matrix).reshape(3, 3)
                elif len(matrix) == 3:
                    matrix = np.diag(matrix)
                else:
                    self.setting_error(
                        "Number of elements of DIM tag has to be 3 or 9."
                    )

                if matrix.shape == (3, 3):
                    if np.linalg.det(matrix) < 1:
                        self.setting_error(
                            "Determinant of supercell matrix has to be positive."
                        )
                    else:
                        self.set_parameter("supercell_matrix", matrix)

            if conf_key in ("primitive_axis", "primitive_axes"):
                if confs[conf_key].strip().lower() == "auto":
                    self.set_parameter("primitive_axes", "auto")
                elif confs[conf_key].strip().upper() in ("P", "F", "I", "A", "C", "R"):
                    self.set_parameter(
                        "primitive_axes", confs[conf_key].strip().upper()
                    )
                elif not len(confs[conf_key].split()) == 9:
                    self.setting_error(
                        "Number of elements in %s has to be 9." % conf_key.upper()
                    )
                else:
                    p_axis = []
                    for x in confs[conf_key].split():
                        p_axis.append(fracval(x))
                    p_axis = np.array(p_axis).reshape(3, 3)
                    if np.linalg.det(p_axis) < 1e-8:
                        self.setting_error(
                            "%s has to have positive determinant." % conf_key.upper()
                        )
                    self.set_parameter("primitive_axes", p_axis)

            if conf_key == "mass":
                self.set_parameter("mass", [float(x) for x in confs["mass"].split()])

            if conf_key == "magmom":
                self.set_parameter(
                    "magmom", [float(x) for x in confs["magmom"].split()]
                )

            if conf_key == "atom_name":
                self.set_parameter(
                    "atom_name", [x.capitalize() for x in confs["atom_name"].split()]
                )

            if conf_key == "displacement_distance":
                self.set_parameter(
                    "displacement_distance", float(confs["displacement_distance"])
                )

            if conf_key == "displacement_distance_max":
                self.set_parameter(
                    "displacement_distance_max",
                    float(confs["displacement_distance_max"]),
                )

            if conf_key == "diag":
                if confs["diag"].lower() == ".false.":
                    self.set_parameter("diag", False)
                elif confs["diag"].lower() == ".true.":
                    self.set_parameter("diag", True)

            if conf_key == "pm":
                if confs["pm"].lower() == ".false.":
                    self.set_parameter("pm", False)
                elif confs["pm"].lower() == ".true.":
                    self.set_parameter("pm", True)

            if conf_key == "trigonal":
                if confs["trigonal"].lower() == ".false.":
                    self.set_parameter("is_trigonal_displacement", False)
                elif confs["trigonal"].lower() == ".true.":
                    self.set_parameter("is_trigonal_displacement", True)

            if conf_key == "eigenvectors":
                if confs["eigenvectors"].lower() == ".false.":
                    self.set_parameter("is_eigenvectors", False)
                elif confs["eigenvectors"].lower() == ".true.":
                    self.set_parameter("is_eigenvectors", True)

            if conf_key == "nac":
                if confs["nac"].lower() == ".false.":
                    self.set_parameter("is_nac", False)
                elif confs["nac"].lower() == ".true.":
                    self.set_parameter("is_nac", True)

            if conf_key == "symmetry":
                if confs["symmetry"].lower() == ".false.":
                    self.set_parameter("is_symmetry", False)
                    self.set_parameter("is_mesh_symmetry", False)
                elif confs["symmetry"].lower() == ".true.":
                    self.set_parameter("is_symmetry", True)

            if conf_key == "mesh_symmetry":
                if confs["mesh_symmetry"].lower() == ".false.":
                    self.set_parameter("is_mesh_symmetry", False)
                elif confs["mesh_symmetry"].lower() == ".true.":
                    self.set_parameter("is_mesh_symmetry", True)

            if conf_key == "calculator":
                self.set_parameter("calculator", confs["calculator"])

            if conf_key == "fc_calculator":
                self.set_parameter("fc_calculator", confs["fc_calculator"])

            if conf_key == "fc_calculator_options":
                self.set_parameter(
                    "fc_calculator_options", confs["fc_calculator_options"]
                )

            if conf_key == "fc_symmetry":
                if confs["fc_symmetry"].lower() == ".false.":
                    self.set_parameter("fc_symmetry", False)
                elif confs["fc_symmetry"].lower() == ".true.":
                    self.set_parameter("fc_symmetry", True)

            if conf_key == "fc_decimals":
                self.set_parameter("fc_decimals", confs["fc_decimals"])

            if conf_key == "dm_decimals":
                self.set_parameter("dm_decimals", confs["dm_decimals"])

            if conf_key in ["mesh_numbers", "mp", "mesh"]:
                vals = [x for x in confs[conf_key].split()]
                if len(vals) == 1:
                    self.set_parameter("mesh_numbers", float(vals[0]))
                elif len(vals) < 3:
                    self.setting_error("Mesh numbers are incorrectly set.")
                elif len(vals) == 3:
                    self.set_parameter("mesh_numbers", [int(x) for x in vals])
                elif len(vals) == 9:
                    mesh_array = []
                    for row in np.reshape([int(x) for x in vals], (3, 3)):
                        mesh_array.append(row.tolist())
                    self.set_parameter("mesh_numbers", mesh_array)
                else:
                    self.setting_error(f"{conf_key.upper()} is incorrectly set.")

            if conf_key == "band_points":
                self.set_parameter("band_points", int(confs["band_points"]))

            if conf_key == "band_const_interval":
                if confs["band_const_interval"].lower() == ".false.":
                    self.set_parameter("is_band_const_interval", False)
                elif confs["band_const_interval"].lower() == ".true.":
                    self.set_parameter("is_band_const_interval", True)

            if conf_key == "band":
                bands = []
                if confs["band"].strip().lower() == "auto":
                    self.set_parameter("band_paths", "auto")
                else:
                    for section in confs["band"].split(","):
                        points = [fracval(x) for x in section.split()]
                        if len(points) % 3 != 0 or len(points) < 6:
                            self.setting_error("BAND is incorrectly set.")
                            break
                        bands.append(np.array(points).reshape(-1, 3))
                    self.set_parameter("band_paths", bands)

            if conf_key == "qpoints":
                if confs["qpoints"].lower() == ".true.":
                    self.set_parameter("read_qpoints", True)
                elif confs["qpoints"].lower() == ".false.":
                    self.set_parameter("read_qpoints", False)
                else:
                    vals = [fracval(x) for x in confs["qpoints"].split()]
                    if len(vals) == 0 or len(vals) % 3 != 0:
                        self.setting_error("Q-points are incorrectly set.")
                    else:
                        self.set_parameter("qpoints", list(np.reshape(vals, (-1, 3))))

            if conf_key == "read_qpoints":
                if confs["read_qpoints"].lower() == ".false.":
                    self.set_parameter("read_qpoints", False)
                elif confs["read_qpoints"].lower() == ".true.":
                    self.set_parameter("read_qpoints", True)

            if conf_key == "nac_method":
                self.set_parameter("nac_method", confs["nac_method"].lower())

            if conf_key == "q_direction":
                q_direction = [fracval(x) for x in confs["q_direction"].split()]
                if len(q_direction) < 3:
                    self.setting_error(
                        "Number of elements of q_direction is less than 3"
                    )
                else:
                    self.set_parameter("nac_q_direction", q_direction)

            if conf_key == "frequency_conversion_factor":
                val = float(confs["frequency_conversion_factor"])
                self.set_parameter("frequency_conversion_factor", val)

            if conf_key == "frequency_scale_factor":
                self.set_parameter(
                    "frequency_scale_factor", float(confs["frequency_scale_factor"])
                )

            if conf_key == "fpitch":
                val = float(confs["fpitch"])
                self.set_parameter("fpitch", val)

            if conf_key == "num_frequency_points":
                val = int(confs["num_frequency_points"])
                self.set_parameter("num_frequency_points", val)

            if conf_key == "cutoff_frequency":
                val = float(confs["cutoff_frequency"])
                self.set_parameter("cutoff_frequency", val)

            if conf_key == "classical":
                if confs["classical"].lower() == ".false.":
                    self.set_parameter("classical", False)
                elif confs["classical"].lower() == ".true.":
                    self.set_parameter("classical", True)

            if conf_key == "sigma":
                vals = [float(x) for x in str(confs["sigma"]).split()]
                if len(vals) == 1:
                    self.set_parameter("sigma", vals[0])
                else:
                    self.set_parameter("sigma", vals)

            if conf_key == "tetrahedron":
                if confs["tetrahedron"].lower() == ".false.":
                    self.set_parameter("is_tetrahedron_method", False)
                if confs["tetrahedron"].lower() == ".true.":
                    self.set_parameter("is_tetrahedron_method", True)

            if conf_key == "symmetry_tolerance":
                val = float(confs["symmetry_tolerance"])
                self.set_parameter("symmetry_tolerance", val)

            if conf_key == "tmin":
                val = float(confs["tmin"])
                self.set_parameter("tmin", val)

            if conf_key == "tmax":
                val = float(confs["tmax"])
                self.set_parameter("tmax", val)

            if conf_key == "tstep":
                val = float(confs["tstep"])
                self.set_parameter("tstep", val)

            # Group velocity finite difference
            if conf_key == "gv_delta_q":
                self.set_parameter("gv_delta_q", float(confs["gv_delta_q"]))

            # Compression option for writing int hdf5
            if conf_key == "hdf5_compression":
                hdf5_compression = confs["hdf5_compression"]
                try:
                    compression = int(hdf5_compression)
                except ValueError:  # str
                    compression = hdf5_compression
                    if compression.lower() == "none":
                        compression = None
                except TypeError:  # None (this will not happen)
                    compression = hdf5_compression
                self.set_parameter("hdf5_compression", compression)

            # Select yaml summary contents
            if conf_key == "save_params":
                if confs["save_params"].lower() == ".true.":
                    self.set_parameter("save_params", True)
                elif confs["save_params"].lower() == ".false.":
                    self.set_parameter("save_params", False)

            if conf_key == "use_pypolymlp":
                if confs["use_pypolymlp"].lower() == ".true.":
                    self.set_parameter("use_pypolymlp", True)
                elif confs["use_pypolymlp"].lower() == ".false.":
                    self.set_parameter("use_pypolymlp", False)

            if conf_key == "mlp_params":
                self.set_parameter("mlp_params", confs["mlp_params"])

            # Number of supercells with random displacements
            if conf_key == "random_displacements":
                rd = confs["random_displacements"]
                if rd.lower() == "auto":
                    self.set_parameter("random_displacements", "auto")
                else:
                    try:
                        self.set_parameter("random_displacements", int(rd))
                    except ValueError:
                        self.setting_error(f"{conf_key.upper()} is incorrectly set.")

            if conf_key == "random_seed":
                self.set_parameter("random_seed", int(confs["random_seed"]))

    def set_parameter(self, key, val):
        """Pass to another data structure."""
        self._parameters[key] = val

    def set_settings(self):
        """Store parameters in Settings class instance."""
        params = self._parameters

        # Chemical symbols
        if "atom_name" in params:
            self._settings.set_chemical_symbols(params["atom_name"])

        # Sets of band indices that are summed
        if "band_indices" in params:
            self._settings.set_band_indices(params["band_indices"])

        # Filename of input unit cell
        if "cell_filename" in params:
            self._settings.set_cell_filename(params["cell_filename"])

        # Is getting least displacements?
        if "create_displacements" in params:
            self._settings.set_create_displacements(params["create_displacements"])

        # Treat statistics classically?
        if "classical" in params:
            self._settings.set_classical(params["classical"])

        # Cutoff frequency
        if "cutoff_frequency" in params:
            self._settings.set_cutoff_frequency(params["cutoff_frequency"])

        # Diagonal displacement
        if "diag" in params:
            self._settings.set_is_diagonal_displacement(params["diag"])

        # Distance of finite displacements introduced
        if "displacement_distance" in params:
            self._settings.set_displacement_distance(params["displacement_distance"])

        if "displacement_distance_max" in params:
            self._settings.set_displacement_distance_max(
                params["displacement_distance_max"]
            )

        # Decimals of values of dynamical matrxi
        if "dm_decimals" in params:
            self._settings.set_dm_decimals(int(params["dm_decimals"]))

        # Force calculator
        if "calculator" in params:
            self._settings.set_calculator(params["calculator"])

        # Force constants calculator
        if "fc_calculator" in params:
            self._settings.set_fc_calculator(params["fc_calculator"])

        # Force constants calculator options as str
        if "fc_calculator_options" in params:
            self._settings.set_fc_calculator_options(params["fc_calculator_options"])

        # Decimals of values of force constants
        if "fc_decimals" in params:
            self._settings.set_fc_decimals(int(params["fc_decimals"]))

        # Enforce translational invariance and index permutation symmetry
        # to force constants?
        if "fc_symmetry" in params:
            self._settings.set_fc_symmetry(params["fc_symmetry"])

        # Frequency unit conversion factor
        if "frequency_conversion_factor" in params:
            self._settings.set_frequency_conversion_factor(
                params["frequency_conversion_factor"]
            )

        # This scale factor is multiplied to force constants by
        # fc * scale_factor ** 2, therefore only changes
        # frequencies but does not change NAC part.
        if "frequency_scale_factor" in params:
            self._settings.set_frequency_scale_factor(params["frequency_scale_factor"])

        # Spectram drawing step
        if "fpitch" in params:
            self._settings.set_frequency_pitch(params["fpitch"])

        # Number of sampling points for spectram drawing
        if "num_frequency_points" in params:
            self._settings.set_num_frequency_points(params["num_frequency_points"])

        # Group velocity finite difference
        if "gv_delta_q" in params:
            self._settings.set_group_velocity_delta_q(params["gv_delta_q"])

        # Mesh sampling numbers
        if "mesh_numbers" in params:
            self._settings.set_mesh_numbers(params["mesh_numbers"])

        # Is getting eigenvectors?
        if "is_eigenvectors" in params:
            self._settings.set_is_eigenvectors(params["is_eigenvectors"])

        # Is reciprocal mesh symmetry searched?
        if "is_mesh_symmetry" in params:
            self._settings.set_is_mesh_symmetry(params["is_mesh_symmetry"])

        # Non analytical term correction?
        if "is_nac" in params:
            self._settings.set_is_nac(params["is_nac"])

        # Is crystal symmetry searched?
        if "is_symmetry" in params:
            self._settings.set_is_symmetry(params["is_symmetry"])

        # Tetrahedron method
        if "is_tetrahedron_method" in params:
            self._settings.set_is_tetrahedron_method(params["is_tetrahedron_method"])

        # Trigonal displacement
        if "is_trigonal_displacement" in params:
            self._settings.set_is_trigonal_displacement(
                params["is_trigonal_displacement"]
            )

        # Magnetic moments
        if "magmom" in params:
            self._settings.set_magnetic_moments(params["magmom"])

        # Atomic mass
        if "mass" in params:
            self._settings.set_masses(params["mass"])

        # Plus minus displacement
        if "pm" in params:
            self._settings.set_is_plusminus_displacement(params["pm"])

        # Primitive cell shape
        if "primitive_axes" in params:
            self._settings.set_primitive_matrix(params["primitive_axes"])

        # Q-points mode
        if "qpoints" in params:
            self._settings.set_qpoints(params["qpoints"])

        if "read_qpoints" in params:
            if params["read_qpoints"]:
                self._settings.set_read_qpoints(params["read_qpoints"])

        # non analytical term correction method
        if "nac_method" in params:
            self._settings.set_nac_method(params["nac_method"])

        # q-direction for non analytical term correction
        if "nac_q_direction" in params:
            self._settings.set_nac_q_direction(params["nac_q_direction"])

        # Smearing width
        if "sigma" in params:
            self._settings.set_sigma(params["sigma"])

        # Symmetry tolerance
        if "symmetry_tolerance" in params:
            self._settings.set_symmetry_tolerance(params["symmetry_tolerance"])

        # Supercell size
        if "supercell_matrix" in params:
            self._settings.set_supercell_matrix(params["supercell_matrix"])

        # Temperatures or temerature range
        if "tmax" in params:
            self._settings.set_max_temperature(params["tmax"])
        if "tmin" in params:
            self._settings.set_min_temperature(params["tmin"])
        if "tstep" in params:
            self._settings.set_temperature_step(params["tstep"])

        # Band paths
        if "band_paths" in params:
            self._settings.set_band_paths(params["band_paths"])

        # This number includes end points
        if "band_points" in params:
            self._settings.set_band_points(params["band_points"])

        if "is_band_const_interval" in params:
            self._settings.set_is_band_const_interval(params["is_band_const_interval"])

        # Compression option for writing int hdf5
        if "hdf5_compression" in params:
            self._settings.set_hdf5_compression(params["hdf5_compression"])

        # Select yaml summary contents
        if "save_params" in params:
            self._settings.set_save_params(params["save_params"])

        # Machine learning potential
        if "use_pypolymlp" in params:
            self._settings.set_use_pypolymlp(params["use_pypolymlp"])

        if "mlp_params" in params:
            self._settings.set_mlp_params(params["mlp_params"])

        # Number of supercells with random displacements
        if "random_displacements" in params:
            self._settings.set_random_displacements(params["random_displacements"])

        if "random_seed" in params:
            self._settings.set_random_seed(params["random_seed"])


#
# For phonopy
#
class PhonopySettings(Settings):
    """Phonopy settings container.

    Basic part is stored in Settings and extended part is stored in this class.

    This works almost like a dictionary.
    Method names without 'set_' and 'get_' and keys of self._v have to be same.

    """

    _default = {
        "anime_band_index": None,
        "anime_amplitude": None,
        "anime_division": None,
        "anime_qpoint": None,
        "anime_shift": None,
        "anime_type": "v_sim",
        "band_format": "yaml",
        "band_labels": None,
        "create_force_sets": None,
        "create_force_sets_zero": None,
        "create_force_constants": None,
        "cutoff_radius": None,
        "dos": None,
        "fc_spg_symmetry": False,
        "fits_Debye_model": False,
        "max_frequency": None,
        "min_frequency": None,
        "irreps_q_point": None,
        "irreps_tolerance": None,
        "is_band_connection": False,
        "is_dos_mode": False,
        "is_full_fc": False,
        "is_group_velocity": False,
        "is_gamma_center": False,
        "is_hdf5": False,
        "is_legacy_plot": False,
        "is_little_cogroup": False,
        "is_moment": False,
        "is_plusminus_displacement": "auto",
        "is_thermal_displacements": False,
        "is_thermal_displacement_matrices": False,
        "is_thermal_distances": False,
        "is_thermal_properties": False,
        "is_projected_thermal_properties": False,
        "include_force_constants": False,
        "include_force_sets": False,
        "include_nac_params": True,
        "include_displacements": False,
        "lapack_solver": False,
        "mesh_shift": None,
        "mesh_format": "yaml",
        "modulation": None,
        "moment_order": None,
        "pdos_indices": None,
        "pretend_real": False,
        "projection_direction": None,
        "qpoints_format": "yaml",
        "random_displacement_temperature": None,
        "read_force_constants": False,
        "readfc_format": "text",
        "run_mode": None,
        "show_irreps": False,
        "sscha_iterations": None,
        "store_dense_svecs": True,
        "thermal_atom_pairs": None,
        "thermal_displacement_matrix_temperatue": None,
        "write_dynamical_matrices": False,
        "write_mesh": True,
        "write_force_constants": False,
        "writefc_format": "text",
        "xyz_projection": False,
    }

    def __init__(self, default=None):
        """Init method."""
        super().__init__()
        self.default.update(PhonopySettings._default.copy())
        if default is not None:
            self.default.update(default)
        self._v = self.default.copy()

    def set_anime_band_index(self, val):
        """Set anime_band_index."""
        self._v["anime_band_index"] = val

    def set_anime_amplitude(self, val):
        """Set anime_amplitude."""
        self._v["anime_amplitude"] = val

    def set_anime_division(self, val):
        """Set anime_division."""
        self._v["anime_division"] = val

    def set_anime_qpoint(self, val):
        """Set anime_qpoint."""
        self._v["anime_qpoint"] = val

    def set_anime_shift(self, val):
        """Set anime_shift."""
        self._v["anime_shift"] = val

    def set_anime_type(self, val):
        """Set anime_type."""
        self._v["anime_type"] = val

    def set_band_format(self, val):
        """Set band_format."""
        self._v["band_format"] = val

    def set_band_labels(self, val):
        """Set band_labels."""
        self._v["band_labels"] = val

    def set_create_force_sets(self, val):
        """Set create_force_sets."""
        self._v["create_force_sets"] = val

    def set_create_force_sets_zero(self, val):
        """Set create_force_sets_zero."""
        self._v["create_force_sets_zero"] = val

    def set_create_force_constants(self, val):
        """Set create_force_constants."""
        self._v["create_force_constants"] = val

    def set_cutoff_radius(self, val):
        """Set cutoff_radius."""
        self._v["cutoff_radius"] = val

    def set_fc_spg_symmetry(self, val):
        """Set fc_spg_symmetry."""
        self._v["fc_spg_symmetry"] = val

    def set_fits_Debye_model(self, val):
        """Set fits_Debye_model."""
        self._v["fits_Debye_model"] = val

    def set_max_frequency(self, val):
        """Set max_frequency."""
        self._v["max_frequency"] = val

    def set_mesh_shift(self, val):
        """Set mesh_shift."""
        self._v["mesh_shift"] = val

    def set_min_frequency(self, val):
        """Set min_frequency."""
        self._v["min_frequency"] = val

    def set_irreps_q_point(self, val):
        """Set irreps_q_point."""
        self._v["irreps_q_point"] = val

    def set_irreps_tolerance(self, val):
        """Set irreps_tolerance."""
        self._v["irreps_tolerance"] = val

    def set_is_band_connection(self, val):
        """Set is_band_connection."""
        self._v["is_band_connection"] = val

    def set_is_dos_mode(self, val):
        """Set is_dos_mode."""
        self._v["is_dos_mode"] = val

    def set_is_full_fc(self, val):
        """Set is_full_fc."""
        self._v["is_full_fc"] = val

    def set_is_gamma_center(self, val):
        """Set is_gamma_center."""
        self._v["is_gamma_center"] = val

    def set_is_group_velocity(self, val):
        """Set is_group_velocity."""
        self._v["is_group_velocity"] = val

    def set_is_hdf5(self, val):
        """Set is_hdf5."""
        self._v["is_hdf5"] = val

    def set_is_legacy_plot(self, val):
        """Set is_legacy_plot."""
        self._v["is_legacy_plot"] = val

    def set_is_little_cogroup(self, val):
        """Set is_little_cogroup."""
        self._v["is_little_cogroup"] = val

    def set_is_moment(self, val):
        """Set is_moment."""
        self._v["is_moment"] = val

    def set_is_projected_thermal_properties(self, val):
        """Set is_projected_thermal_properties."""
        self._v["is_projected_thermal_properties"] = val

    def set_is_thermal_displacements(self, val):
        """Set is_thermal_displacements."""
        self._v["is_thermal_displacements"] = val

    def set_is_thermal_displacement_matrices(self, val):
        """Set is_thermal_displacement_matrices."""
        self._v["is_thermal_displacement_matrices"] = val

    def set_is_thermal_distances(self, val):
        """Set is_thermal_distances."""
        self._v["is_thermal_distances"] = val

    def set_is_thermal_properties(self, val):
        """Set is_thermal_properties."""
        self._v["is_thermal_properties"] = val

    def set_include_force_constants(self, val):
        """Set include_force_constants."""
        self._v["include_force_constants"] = val

    def set_include_force_sets(self, val):
        """Set include_force_sets."""
        self._v["include_force_sets"] = val

    def set_include_nac_params(self, val):
        """Set include_nac_params."""
        self._v["include_nac_params"] = val

    def set_include_displacements(self, val):
        """Set include_displacements."""
        self._v["include_displacements"] = val

    def set_lapack_solver(self, val):
        """Set lapack_solver."""
        self._v["lapack_solver"] = val

    def set_mesh_format(self, val):
        """Set mesh_format."""
        self._v["mesh_format"] = val

    def set_modulation(self, val):
        """Set modulation."""
        self._v["modulation"] = val

    def set_moment_order(self, val):
        """Set moment_order."""
        self._v["moment_order"] = val

    def set_pdos_indices(self, val):
        """Set pdos_indices."""
        self._v["pdos_indices"] = val

    def set_pretend_real(self, val):
        """Set pretend_real."""
        self._v["pretend_real"] = val

    def set_projection_direction(self, val):
        """Set projection_direction."""
        self._v["projection_direction"] = val

    def set_qpoints_format(self, val):
        """Set qpoints_format."""
        self._v["qpoints_format"] = val

    def set_random_displacement_temperature(self, val):
        """Set random_displacement_temperature."""
        self._v["random_displacement_temperature"] = val

    def set_read_force_constants(self, val):
        """Set read_force_constants."""
        self._v["read_force_constants"] = val

    def set_readfc_format(self, val):
        """Set readfc_format."""
        self._v["readfc_format"] = val

    def set_run_mode(self, val):
        """Set run_mode."""
        self._v["run_mode"] = val

    def set_thermal_atom_pairs(self, val):
        """Set thermal_atom_pairs."""
        self._v["thermal_atom_pairs"] = val

    def set_thermal_displacement_matrix_temperature(self, val):
        """Set thermal_displacement_matrix_temperatue."""
        self._v["thermal_displacement_matrix_temperatue"] = val

    def set_show_irreps(self, val):
        """Set show_irreps."""
        self._v["show_irreps"] = val

    def set_sscha_iterations(self, val):
        """Set sscha_iterations."""
        self._v["sscha_iterations"] = val

    def set_store_dense_svecs(self, val):
        """Set store_dense_svecs."""
        self._v["store_dense_svecs"] = val

    def set_write_dynamical_matrices(self, val):
        """Set write_dynamical_matrices."""
        self._v["write_dynamical_matrices"] = val

    def set_write_force_constants(self, val):
        """Set write_force_constants."""
        self._v["write_force_constants"] = val

    def set_write_mesh(self, val):
        """Set write_mesh."""
        self._v["write_mesh"] = val

    def set_writefc_format(self, val):
        """Set writefc_format."""
        self._v["writefc_format"] = val

    def set_xyz_projection(self, val):
        """Set xyz_projection."""
        self._v["xyz_projection"] = val


class PhonopyConfParser(ConfParser):
    """Phonopy conf parser."""

    def __init__(self, filename=None, args=None, default_settings=None):
        """Init method.

        Note
        ----
        This implementation is fragile.
        Remember that options have to be activated only
        when it changes the default value, i.e.,
        _read_options has to be written in this way.

        """
        self._settings = PhonopySettings(default=default_settings)
        confs = {}
        if filename is not None:
            super().__init__(filename=filename)
            self.read_file()  # store .conf file setting in self._confs
            self._parse_conf()  # self.parameters[key] = val
            self._set_settings()  # self.parameters -> PhonopySettings
            confs.update(self._confs)
        if args is not None:
            # To invoke ConfParser.__init__() to flush variables.
            super().__init__(args=args)
            self._read_options()  # store options in self._confs
            self._parse_conf()  # self.parameters[key] = val
            self._set_settings()  # self.parameters -> PhonopySettings
            confs.update(self._confs)
        self._confs = confs

    def _read_options(self):
        ConfParser.read_options(self)  # store data in self._confs
        arg_list = vars(self._args)
        if "band_format" in arg_list:
            if self._args.band_format:
                self._confs["band_format"] = self._args.band_format

        if "band_labels" in arg_list:
            if self._args.band_labels is not None:
                self._confs["band_labels"] = " ".join(self._args.band_labels)

        if "is_gamma_center" in arg_list:
            if self._args.is_gamma_center:
                self._confs["gamma_center"] = ".true."

        if "create_force_sets" in arg_list:
            if self._args.create_force_sets:
                self._confs["create_force_sets"] = self._args.create_force_sets

        if "create_force_sets_zero" in arg_list:
            if self._args.create_force_sets_zero:
                fc_sets_zero = self._args.create_force_sets_zero
                self._confs["create_force_sets_zero"] = fc_sets_zero

        if "create_force_constants" in arg_list:
            if self._args.create_force_constants is not None:
                fc_filename = self._args.create_force_constants
                self._confs["create_force_constants"] = fc_filename

        if "is_dos_mode" in arg_list:
            if self._args.is_dos_mode:
                self._confs["dos"] = ".true."

        if "pdos" in arg_list:
            if self._args.pdos is not None:
                self._confs["pdos"] = " ".join(self._args.pdos)

        if "xyz_projection" in arg_list:
            if self._args.xyz_projection:
                self._confs["xyz_projection"] = ".true."

        if "fc_spg_symmetry" in arg_list:
            if self._args.fc_spg_symmetry:
                self._confs["fc_spg_symmetry"] = ".true."

        if "is_full_fc" in arg_list:
            if self._args.is_full_fc:
                self._confs["full_force_constants"] = ".true."

        if "fits_debye_model" in arg_list:
            if self._args.fits_debye_model:
                self._confs["debye_model"] = ".true."

        if "fmax" in arg_list:
            if self._args.fmax is not None:
                self._confs["fmax"] = self._args.fmax

        if "fmin" in arg_list:
            if self._args.fmin is not None:
                self._confs["fmin"] = self._args.fmin

        if "is_thermal_properties" in arg_list:
            if self._args.is_thermal_properties:
                self._confs["tprop"] = ".true."

        if "pretend_real" in arg_list:
            if self._args.pretend_real:
                self._confs["pretend_real"] = ".true."

        if "is_projected_thermal_properties" in arg_list:
            if self._args.is_projected_thermal_properties:
                self._confs["ptprop"] = ".true."

        if "is_thermal_displacements" in arg_list:
            if self._args.is_thermal_displacements:
                self._confs["tdisp"] = ".true."

        if "is_thermal_displacement_matrices" in arg_list:
            if self._args.is_thermal_displacement_matrices:
                self._confs["tdispmat"] = ".true."

        if "thermal_displacement_matrices_cif" in arg_list:
            opt_tdm_cif = self._args.thermal_displacement_matrices_cif
            if opt_tdm_cif:
                self._confs["tdispmat_cif"] = opt_tdm_cif

        if "projection_direction" in arg_list:
            opt_proj_dir = self._args.projection_direction
            if opt_proj_dir is not None:
                self._confs["projection_direction"] = " ".join(opt_proj_dir)

        if "read_force_constants" in arg_list:
            if self._args.read_force_constants:
                self._confs["read_force_constants"] = ".true."

        if "write_force_constants" in arg_list:
            if self._args.write_force_constants:
                self._confs["write_force_constants"] = ".true."

        if "readfc_format" in arg_list:
            if self._args.readfc_format:
                self._confs["readfc_format"] = self._args.readfc_format

        if "writefc_format" in arg_list:
            if self._args.writefc_format:
                self._confs["writefc_format"] = self._args.writefc_format

        if "fc_format" in arg_list:
            if self._args.fc_format:
                self._confs["fc_format"] = self._args.fc_format

        if "is_hdf5" in arg_list:
            if self._args.is_hdf5:
                self._confs["hdf5"] = ".true."

        if "write_dynamical_matrices" in arg_list:
            if self._args.write_dynamical_matrices:
                self._confs["writedm"] = ".true."

        if "write_mesh" in arg_list:
            if self._args.write_mesh is False:
                self._confs["write_mesh"] = ".false."

        if "mesh_format" in arg_list:
            if self._args.mesh_format:
                self._confs["mesh_format"] = self._args.mesh_format

        if "qpoints_format" in arg_list:
            if self._args.qpoints_format:
                self._confs["qpoints_format"] = self._args.qpoints_format

        if "irreps_qpoint" in arg_list:
            if self._args.irreps_qpoint is not None:
                self._confs["irreps"] = " ".join(self._args.irreps_qpoint)

        if "show_irreps" in arg_list:
            if self._args.show_irreps:
                self._confs["show_irreps"] = ".true."

        if "is_little_cogroup" in arg_list:
            if self._args.is_little_cogroup:
                self._confs["little_cogroup"] = ".true."

        if "is_legacy_plot" in arg_list:
            if self._args.is_legacy_plot:
                self._confs["legacy_plot"] = ".true."

        if "is_band_connection" in arg_list:
            if self._args.is_band_connection:
                self._confs["band_connection"] = ".true."

        if "cutoff_radius" in arg_list:
            if self._args.cutoff_radius:
                self._confs["cutoff_radius"] = self._args.cutoff_radius

        if "modulation" in arg_list:
            if self._args.modulation:
                self._confs["modulation"] = " ".join(self._args.modulation)

        if "anime" in arg_list:
            if self._args.anime:
                self._confs["anime"] = " ".join(self._args.anime)

        if "is_group_velocity" in arg_list:
            if self._args.is_group_velocity:
                self._confs["group_velocity"] = ".true."

        if "is_moment" in arg_list:
            if self._args.is_moment:
                self._confs["moment"] = ".true."

        if "moment_order" in arg_list:
            if self._args.moment_order:
                self._confs["moment_order"] = self._args.moment_order

        if "rd_temperature" in arg_list:
            if self._args.rd_temperature is not None:
                self._confs["random_displacement_temperature"] = (
                    self._args.rd_temperature
                )

        if "temperature" in arg_list:
            if self._args.temperature is not None:
                print(
                    "*****************************************************************"
                )
                print(
                    "--temperature option is deprecated. Use --rd-temperature instead."
                )
                print(
                    "*****************************************************************"
                )
                print()
                self._confs["random_displacement_temperature"] = self._args.temperature

        if "include_fc" in arg_list:
            if self._args.include_fc:
                self._confs["include_fc"] = ".true."

        if "include_fs" in arg_list:
            if self._args.include_fs:
                self._confs["include_fs"] = ".true."

        if "include_nac_params" in arg_list:
            if self._settings.default["include_nac_params"]:
                if self._args.include_nac_params is False:
                    self._confs["include_nac_params"] = ".false."
            else:
                if self._args.include_nac_params:
                    self._confs["include_nac_params"] = ".true."

        if "include_disp" in arg_list:
            if self._args.include_disp:
                self._confs["include_disp"] = ".true."

        if "include_all" in arg_list:
            if self._args.include_all:
                self._confs["include_all"] = ".true."

        if "store_dense_svecs" in arg_list:
            if self._args.store_dense_svecs:
                self._confs["store_dense_svecs"] = ".true."
            else:
                self._confs["store_dense_svecs"] = ".false."

        if "lapack_solver" in arg_list:
            if self._args.lapack_solver:
                self._confs["lapack_solver"] = ".true."

        if "is_check_symmetry" in arg_list:
            if self._args.is_check_symmetry:
                if "dim" not in self._confs:
                    # Dummy 'dim' setting not to exit by no-dim check.
                    self._confs["dim"] = "1 1 1"

        if "sscha_iterations" in arg_list:
            if self._args.sscha_iterations:
                self._confs["sscha_iterations"] = self._args.sscha_iterations

    def _parse_conf(self):
        ConfParser.parse_conf(self)
        confs = self._confs

        for conf_key in confs.keys():
            if conf_key == "band_format":
                self.set_parameter("band_format", confs["band_format"].lower())

            if conf_key == "band_labels":
                labels = confs["band_labels"].split()
                self.set_parameter("band_labels", labels)

            if conf_key == "band_connection":
                if confs["band_connection"].lower() == ".true.":
                    self.set_parameter("band_connection", True)
                elif confs["band_connection"].lower() == ".false.":
                    self.set_parameter("band_connection", False)

            if conf_key == "legacy_plot":
                if confs["legacy_plot"].lower() == ".true.":
                    self.set_parameter("legacy_plot", True)
                elif confs["legacy_plot"].lower() == ".false.":
                    self.set_parameter("legacy_plot", False)

            if conf_key == "create_force_sets":
                if isinstance(confs["create_force_sets"], str):
                    fnames = confs["create_force_sets"].split()
                else:
                    fnames = confs["create_force_sets"]
                self.set_parameter("create_force_sets", fnames)

            if conf_key == "create_force_sets_zero":
                if isinstance(confs["create_force_sets_zero"], str):
                    fnames = confs["create_force_sets_zero"].split()
                else:
                    fnames = confs["create_force_sets_zero"]

                self.set_parameter("create_force_sets_zero", fnames)

            if conf_key == "create_force_constants":
                self.set_parameter(
                    "create_force_constants", confs["create_force_constants"]
                )

            if conf_key == "force_constants":
                self.set_parameter("force_constants", confs["force_constants"].lower())

            if conf_key == "read_force_constants":
                if confs["read_force_constants"].lower() == ".true.":
                    self.set_parameter("read_force_constants", True)
                elif confs["read_force_constants"].lower() == ".false.":
                    self.set_parameter("read_force_constants", False)

            if conf_key == "write_force_constants":
                if confs["write_force_constants"].lower() == ".true.":
                    self.set_parameter("write_force_constants", True)
                elif confs["write_force_constants"].lower() == ".false.":
                    self.set_parameter("write_force_constants", False)

            if conf_key == "full_force_constants":
                if confs["full_force_constants"].lower() == ".true.":
                    self.set_parameter("is_full_fc", True)
                elif confs["full_force_constants"].lower() == ".false.":
                    self.set_parameter("is_full_fc", False)

            if conf_key == "cutoff_radius":
                val = float(confs["cutoff_radius"])
                self.set_parameter("cutoff_radius", val)

            if conf_key == "writedm":
                if confs["writedm"].lower() == ".true.":
                    self.set_parameter("write_dynamical_matrices", True)
                elif confs["writedm"].lower() == ".false.":
                    self.set_parameter("write_dynamical_matrices", False)

            if conf_key == "write_mesh":
                if confs["write_mesh"].lower() == ".true.":
                    self.set_parameter("write_mesh", True)
                elif confs["write_mesh"].lower() == ".false.":
                    self.set_parameter("write_mesh", False)

            if conf_key == "hdf5":
                if confs["hdf5"].lower() == ".true.":
                    self.set_parameter("hdf5", True)
                elif confs["hdf5"].lower() == ".false.":
                    self.set_parameter("hdf5", False)

            if conf_key == "mp_shift":
                vals = [fracval(x) for x in confs["mp_shift"].split()]
                if len(vals) < 3:
                    self.setting_error("MP_SHIFT is incorrectly set.")
                self.set_parameter("mp_shift", vals[:3])

            if conf_key == "mesh_format":
                self.set_parameter("mesh_format", confs["mesh_format"].lower())

            if conf_key == "qpoints_format":
                self.set_parameter("qpoints_format", confs["qpoints_format"].lower())

            if conf_key == "time_reversal_symmetry":
                if confs["time_reversal_symmetry"].lower() == ".true.":
                    self.set_parameter("is_time_reversal_symmetry", True)
                elif confs["time_reversal_symmetry"].lower() == ".false.":
                    self.set_parameter("is_time_reversal_symmetry", False)

            if conf_key == "gamma_center":
                if confs["gamma_center"].lower() == ".true.":
                    self.set_parameter("is_gamma_center", True)
                elif confs["gamma_center"].lower() == ".false.":
                    self.set_parameter("is_gamma_center", False)

            if conf_key == "fc_spg_symmetry":
                if confs["fc_spg_symmetry"].lower() == ".true.":
                    self.set_parameter("fc_spg_symmetry", True)
                elif confs["fc_spg_symmetry"].lower() == ".false.":
                    self.set_parameter("fc_spg_symmetry", False)

            if conf_key == "readfc_format":
                self.set_parameter("readfc_format", confs["readfc_format"].lower())

            if conf_key == "writefc_format":
                self.set_parameter("writefc_format", confs["writefc_format"].lower())

            if conf_key == "fc_format":
                self.set_parameter("readfc_format", confs["fc_format"].lower())
                self.set_parameter("writefc_format", confs["fc_format"].lower())

            # Animation
            if conf_key == "anime":
                vals = []
                data = confs["anime"].split()
                if len(data) < 3:
                    self.setting_error("ANIME is incorrectly set.")
                else:
                    self.set_parameter("anime", data)

            if conf_key == "anime_type":
                anime_type = confs["anime_type"].lower()
                if anime_type in ("arc", "v_sim", "poscar", "xyz", "jmol"):
                    self.set_parameter("anime_type", anime_type)
                else:
                    self.setting_error(
                        "%s is not available for ANIME_TYPE tag." % confs["anime_type"]
                    )

            # Modulation
            if conf_key == "modulation":
                self._parse_conf_modulation(confs["modulation"])

            # Character table
            if conf_key == "irreps":
                vals = [fracval(x) for x in confs["irreps"].split()]
                if len(vals) == 3 or len(vals) == 4:
                    self.set_parameter("irreps_qpoint", vals)
                else:
                    self.setting_error("IRREPS is incorrectly set.")

            if conf_key == "show_irreps":
                if confs["show_irreps"].lower() == ".true.":
                    self.set_parameter("show_irreps", True)
                elif confs["show_irreps"].lower() == ".false.":
                    self.set_parameter("show_irreps", False)

            if conf_key == "little_cogroup":
                if confs["little_cogroup"].lower() == ".true.":
                    self.set_parameter("little_cogroup", True)
                elif confs["little_cogroup"].lower() == ".false.":
                    self.set_parameter("little_cogroup", False)

            # DOS
            if conf_key == "pdos":
                if confs["pdos"].strip().lower() == "auto":
                    self.set_parameter("pdos", "auto")
                else:
                    vals = []
                    for index_set in confs["pdos"].split(","):
                        vals.append([int(x) - 1 for x in index_set.split()])
                    self.set_parameter("pdos", vals)

            if conf_key == "xyz_projection":
                if confs["xyz_projection"].lower() == ".true.":
                    self.set_parameter("xyz_projection", True)
                elif confs["xyz_projection"].lower() == ".false.":
                    self.set_parameter("xyz_projection", False)

            if conf_key == "dos":
                if confs["dos"].lower() == ".true.":
                    self.set_parameter("dos", True)
                elif confs["dos"].lower() == ".false.":
                    self.set_parameter("dos", False)

            if conf_key == "debye_model":
                if confs["debye_model"].lower() == ".true.":
                    self.set_parameter("fits_debye_model", True)
                elif confs["debye_model"].lower() == ".false.":
                    self.set_parameter("fits_debye_model", False)

            if conf_key == "dos_range":
                vals = [float(x) for x in confs["dos_range"].split()]
                self.set_parameter("dos_range", vals)

            if conf_key == "fmax":
                self.set_parameter("fmax", float(confs["fmax"]))

            if conf_key == "fmin":
                self.set_parameter("fmin", float(confs["fmin"]))

            # Thermal properties
            if conf_key == "tprop":
                if confs["tprop"].lower() == ".true.":
                    self.set_parameter("tprop", True)
                if confs["tprop"].lower() == ".false.":
                    self.set_parameter("tprop", False)

            # Projected thermal properties
            if conf_key == "ptprop":
                if confs["ptprop"].lower() == ".true.":
                    self.set_parameter("ptprop", True)
                elif confs["ptprop"].lower() == ".false.":
                    self.set_parameter("ptprop", False)

            # Use imaginary frequency as real for thermal property calculation
            if conf_key == "pretend_real":
                if confs["pretend_real"].lower() == ".true.":
                    self.set_parameter("pretend_real", True)
                elif confs["pretend_real"].lower() == ".false.":
                    self.set_parameter("pretend_real", False)

            # Thermal displacement
            if conf_key == "tdisp":
                if confs["tdisp"].lower() == ".true.":
                    self.set_parameter("tdisp", True)
                elif confs["tdisp"].lower() == ".false.":
                    self.set_parameter("tdisp", False)

            # Thermal displacement matrices
            if conf_key == "tdispmat":
                if confs["tdispmat"].lower() == ".true.":
                    self.set_parameter("tdispmat", True)
                elif confs["tdispmat"].lower() == ".false.":
                    self.set_parameter("tdispmat", False)

            # Write thermal displacement matrices to cif file,
            # for which the temperature to execute is stored.
            if conf_key == "tdispmat_cif":
                self.set_parameter("tdispmat_cif", float(confs["tdispmat_cif"]))

            # Thermal distance
            if conf_key == "tdistance":
                atom_pairs = []
                for atoms in confs["tdistance"].split(","):
                    pair = [int(x) - 1 for x in atoms.split()]
                    if len(pair) == 2:
                        atom_pairs.append(pair)
                    else:
                        self.setting_error("TDISTANCE is incorrectly specified.")
                if len(atom_pairs) > 0:
                    self.set_parameter("tdistance", atom_pairs)

            # Projection direction used for thermal displacements and PDOS
            if conf_key == "projection_direction":
                vals = [float(x) for x in confs["projection_direction"].split()]
                if len(vals) < 3:
                    self.setting_error(
                        "PROJECTION_DIRECTION (--pd) is incorrectly specified."
                    )
                else:
                    self.set_parameter("projection_direction", vals)

            # Group velocity
            if conf_key == "group_velocity":
                if confs["group_velocity"].lower() == ".true.":
                    self.set_parameter("is_group_velocity", True)
                elif confs["group_velocity"].lower() == ".false.":
                    self.set_parameter("is_group_velocity", False)

            # Moment of phonon states distribution
            if conf_key == "moment":
                if confs["moment"].lower() == ".true.":
                    self.set_parameter("moment", True)
                elif confs["moment"].lower() == ".false.":
                    self.set_parameter("moment", False)

            if conf_key == "moment_order":
                self.set_parameter("moment_order", int(confs["moment_order"]))

            if conf_key == "random_displacement_temperature":
                val = confs["random_displacement_temperature"]
                self.set_parameter("random_displacement_temperature", float(val))

            # Use Lapack solver via Lapacke
            if conf_key == "lapack_solver":
                if confs["lapack_solver"].lower() == ".true.":
                    self.set_parameter("lapack_solver", True)
                elif confs["lapack_solver"].lower() == ".false.":
                    self.set_parameter("lapack_solver", False)

            if conf_key == "include_fc":
                if confs["include_fc"].lower() == ".true.":
                    self.set_parameter("include_fc", True)
                elif confs["include_fc"].lower() == ".false.":
                    self.set_parameter("include_fc", False)

            if conf_key == "include_fs":
                if confs["include_fs"].lower() == ".true.":
                    self.set_parameter("include_fs", True)
                elif confs["include_fs"].lower() == ".false.":
                    self.set_parameter("include_fs", False)

            if conf_key in ("include_born", "include_nac_params"):
                if confs[conf_key].lower() == ".true.":
                    self.set_parameter("include_nac_params", True)
                elif confs[conf_key].lower() == ".false.":
                    self.set_parameter("include_nac_params", False)

            if conf_key == "include_disp":
                if confs["include_disp"].lower() == ".true.":
                    self.set_parameter("include_disp", True)
                elif confs["include_disp"].lower() == ".false.":
                    self.set_parameter("include_disp", False)

            if conf_key == "include_all":
                if confs["include_all"].lower() == ".true.":
                    self.set_parameter("include_all", True)
                elif confs["include_all"].lower() == ".false.":
                    self.set_parameter("include_all", False)

            # Pair shortest vectors in supercell are stored in dense format.
            if conf_key == "store_dense_svecs":
                if confs["store_dense_svecs"].lower() == ".true.":
                    self.set_parameter("store_dense_svecs", True)

            # SSCHA
            if conf_key == "sscha_iterations":
                val = int(confs["sscha_iterations"])
                self.set_parameter("sscha_iterations", val)

    def _parse_conf_modulation(self, conf_modulation):
        modulation = {}
        modulation["dimension"] = [1, 1, 1]
        modulation["order"] = None
        mod_list = conf_modulation.split(",")
        header = mod_list[0].split()
        if len(header) > 2 and len(mod_list) > 1:
            if len(header) > 8:
                dimension = [int(x) for x in header[:9]]
                modulation["dimension"] = dimension
                if len(header) > 11:
                    delta_q = [float(x) for x in header[9:12]]
                    modulation["delta_q"] = delta_q
                if len(header) == 13:
                    modulation["order"] = int(header[12])
            else:
                dimension = [int(x) for x in header[:3]]
                modulation["dimension"] = dimension
                if len(header) > 3:
                    delta_q = [float(x) for x in header[3:6]]
                    modulation["delta_q"] = delta_q
                if len(header) == 7:
                    modulation["order"] = int(header[6])

            vals = []
            for phonon_mode in mod_list[1:]:
                mode_conf = [x for x in phonon_mode.split()]
                if len(mode_conf) < 4 or len(mode_conf) > 6:
                    self.setting_error("MODULATION tag is wrongly set.")
                    break
                else:
                    q = [fracval(x) for x in mode_conf[:3]]

                if len(mode_conf) == 4:
                    vals.append([q, int(mode_conf[3]) - 1, 1.0, 0])
                elif len(mode_conf) == 5:
                    vals.append([q, int(mode_conf[3]) - 1, float(mode_conf[4]), 0])
                else:
                    vals.append(
                        [
                            q,
                            int(mode_conf[3]) - 1,
                            float(mode_conf[4]),
                            float(mode_conf[5]),
                        ]
                    )

            modulation["modulations"] = vals
            self.set_parameter("modulation", modulation)
        else:
            self.setting_error("MODULATION tag is wrongly set.")

    def _set_settings(self):
        self.set_settings()
        params = self._parameters

        # Create FORCE_SETS
        if "create_force_sets" in params:
            self._settings.set_create_force_sets(params["create_force_sets"])

        if "create_force_sets_zero" in params:
            self._settings.set_create_force_sets_zero(params["create_force_sets_zero"])

        if "create_force_constants" in params:
            self._settings.set_create_force_constants(params["create_force_constants"])

        # Is force constants written or read?
        if "force_constants" in params:
            if params["force_constants"] == "write":
                self._settings.set_write_force_constants(True)
            elif params["force_constants"] == "read":
                self._settings.set_read_force_constants(True)

        if "read_force_constants" in params:
            self._settings.set_read_force_constants(params["read_force_constants"])

        if "write_force_constants" in params:
            self._settings.set_write_force_constants(params["write_force_constants"])

        if "is_full_fc" in params:
            self._settings.set_is_full_fc(params["is_full_fc"])

        # Enforce space group symmetyr to force constants?
        if "fc_spg_symmetry" in params:
            self._settings.set_fc_spg_symmetry(params["fc_spg_symmetry"])

        if "readfc_format" in params:
            self._settings.set_readfc_format(params["readfc_format"])

        if "writefc_format" in params:
            self._settings.set_writefc_format(params["writefc_format"])

        # Use hdf5?
        if "hdf5" in params:
            self._settings.set_is_hdf5(params["hdf5"])

        # Cutoff radius of force constants
        if "cutoff_radius" in params:
            self._settings.set_cutoff_radius(params["cutoff_radius"])

        # Mesh
        if "mesh_numbers" in params:
            self._settings.set_run_mode("mesh")
            self._settings.set_mesh_numbers(params["mesh_numbers"])
        if "mp_shift" in params:
            self._settings.set_mesh_shift(params["mp_shift"])
        if "is_time_reversal_symmetry" in params:
            self._settings.set_is_time_reversal_symmetry(
                params["is_time_reversal_symmetry"]
            )
        if "is_mesh_symmetry" in params:
            self._settings.set_is_mesh_symmetry(params["is_mesh_symmetry"])
        if "is_gamma_center" in params:
            self._settings.set_is_gamma_center(params["is_gamma_center"])
        if "mesh_format" in params:
            self._settings.set_mesh_format(params["mesh_format"])

        # band mode
        if "band_paths" in params:
            self._settings.set_run_mode("band")
        if "band_format" in params:
            self._settings.set_band_format(params["band_format"])
        if "band_labels" in params:
            self._settings.set_band_labels(params["band_labels"])
        if "band_connection" in params:
            self._settings.set_is_band_connection(params["band_connection"])
        if "legacy_plot" in params:
            self._settings.set_is_legacy_plot(params["legacy_plot"])

        # Q-points mode
        if "qpoints" in params or "read_qpoints" in params:
            self._settings.set_run_mode("qpoints")
        if self._settings.run_mode == "qpoints":
            if "qpoints_format" in params:
                self._settings.set_qpoints_format(params["qpoints_format"])

        # Whether write out dynamical matrices or not
        if "write_dynamical_matrices" in params:
            self._settings.set_write_dynamical_matrices(
                params["write_dynamical_matrices"]
            )

        # Whether write out mesh.yaml or mesh.hdf5
        if "write_mesh" in params:
            self._settings.set_write_mesh(params["write_mesh"])

        # Anime mode
        if "anime_type" in params:
            self._settings.set_anime_type(params["anime_type"])

        if "anime" in params:
            self._settings.set_run_mode("anime")
            anime_type = self._settings.anime_type
            if anime_type == "v_sim":
                qpoints = [fracval(x) for x in params["anime"][0:3]]
                self._settings.set_anime_qpoint(qpoints)
                if len(params["anime"]) > 3:
                    self._settings.set_anime_amplitude(float(params["anime"][3]))
            else:
                self._settings.set_anime_band_index(int(params["anime"][0]))
                self._settings.set_anime_amplitude(float(params["anime"][1]))
                self._settings.set_anime_division(int(params["anime"][2]))
            if len(params["anime"]) == 6:
                self._settings.set_anime_shift(
                    [fracval(x) for x in params["anime"][3:6]]
                )

        # Modulation mode
        if "modulation" in params:
            self._settings.set_run_mode("modulation")
            self._settings.set_modulation(params["modulation"])

        # Character table mode
        if "irreps_qpoint" in params:
            self._settings.set_run_mode("irreps")
            self._settings.set_irreps_q_point(params["irreps_qpoint"][:3])
            if len(params["irreps_qpoint"]) == 4:
                self._settings.set_irreps_tolerance(params["irreps_qpoint"][3])
        if "show_irreps" in params:
            self._settings.set_show_irreps(params["show_irreps"])
        if "little_cogroup" in params:
            self._settings.set_is_little_cogroup(params["little_cogroup"])

        # DOS
        if "dos_range" in params:
            fmin = params["dos_range"][0]
            fmax = params["dos_range"][1]
            fpitch = params["dos_range"][2]
            self._settings.set_min_frequency(fmin)
            self._settings.set_max_frequency(fmax)
            self._settings.set_frequency_pitch(fpitch)
        if "dos" in params:
            self._settings.set_is_dos_mode(params["dos"])

        if "fits_debye_model" in params:
            self._settings.set_fits_Debye_model(params["fits_debye_model"])

        if "fmax" in params:
            self._settings.set_max_frequency(params["fmax"])

        if "fmin" in params:
            self._settings.set_min_frequency(params["fmin"])

        # Project PDOS x, y, z directions in Cartesian coordinates
        if "xyz_projection" in params:
            self._settings.set_xyz_projection(params["xyz_projection"])
            if "pdos" not in params and self._settings.pdos_indices is None:
                self.set_parameter("pdos", [])

        if "pdos" in params:
            self._settings.set_pdos_indices(params["pdos"])
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)

        if "projection_direction" in params and not self._settings.xyz_projection:
            self._settings.set_projection_direction(params["projection_direction"])
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)

        # Thermal properties
        if "tprop" in params:
            self._settings.set_is_thermal_properties(params["tprop"])
            # Exclusive conditions
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_displacement_matrices(False)
            self._settings.set_is_thermal_distances(False)

        # Projected thermal properties
        if "ptprop" in params and params["ptprop"]:
            self._settings.set_is_thermal_properties(True)
            self._settings.set_is_projected_thermal_properties(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            # Exclusive conditions
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_displacement_matrices(False)
            self._settings.set_is_thermal_distances(False)

        # Use imaginary frequency as real for thermal property calculation
        if "pretend_real" in params:
            self._settings.set_pretend_real(params["pretend_real"])

        # Thermal displacements
        if "tdisp" in params and params["tdisp"]:
            self._settings.set_is_thermal_displacements(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            # Exclusive conditions
            self._settings.set_is_thermal_properties(False)
            self._settings.set_is_thermal_displacement_matrices(False)
            self._settings.set_is_thermal_distances(True)

        # Thermal displacement matrices
        if "tdispmat" in params and params["tdispmat"] or "tdispmat_cif" in params:
            self._settings.set_is_thermal_displacement_matrices(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            # Exclusive conditions
            self._settings.set_is_thermal_properties(False)
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_distances(False)

            # Temperature used to calculate thermal displacement matrix
            # to write aniso_U to cif
            if "tdispmat_cif" in params:
                self._settings.set_thermal_displacement_matrix_temperature(
                    params["tdispmat_cif"]
                )

        # Thermal distances
        if "tdistance" in params:
            self._settings.set_is_thermal_distances(True)
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)
            self._settings.set_thermal_atom_pairs(params["tdistance"])
            # Exclusive conditions
            self._settings.set_is_thermal_properties(False)
            self._settings.set_is_thermal_displacements(False)
            self._settings.set_is_thermal_displacement_matrices(False)

        # Group velocity
        if "is_group_velocity" in params:
            self._settings.set_is_group_velocity(params["is_group_velocity"])

        # Moment mode
        if "moment" in params:
            self._settings.set_is_moment(params["moment"])
            self._settings.set_is_eigenvectors(True)
            self._settings.set_is_mesh_symmetry(False)

        if self._settings.is_moment:
            if "moment_order" in params:
                self._settings.set_moment_order(params["moment_order"])

        if "random_displacement_temperature" in params:
            self._settings.set_random_displacement_temperature(
                params["random_displacement_temperature"]
            )

        # Use Lapack solver via Lapacke
        if "lapack_solver" in params:
            self._settings.set_lapack_solver(params["lapack_solver"])

        if "include_fc" in params:
            self._settings.set_include_force_constants(params["include_fc"])

        if "include_fs" in params:
            self._settings.set_include_force_sets(params["include_fs"])

        if "include_nac_params" in params:
            self._settings.set_include_nac_params(params["include_nac_params"])

        if "include_disp" in params:
            self._settings.set_include_displacements(params["include_disp"])

        if "include_all" in params:
            self._settings.set_include_force_constants(True)
            self._settings.set_include_force_sets(True)
            self._settings.set_include_nac_params(True)
            self._settings.set_include_displacements(True)

        # Pair shortest vectors in supercell are stored in dense format.
        if "store_dense_svecs" in params:
            self._settings.set_store_dense_svecs(params["store_dense_svecs"])

        # ***********************************************************
        # This has to come last in this method to overwrite run_mode.
        # ***********************************************************
        if "pdos" in params and params["pdos"] == "auto":
            if "band_paths" in params:
                self._settings.set_run_mode("band_mesh")
            else:
                self._settings.set_run_mode("mesh")

        if "mesh_numbers" in params and "band_paths" in params:
            self._settings.set_run_mode("band_mesh")

        # SSCHA
        if "sscha_iterations" in params:
            self._settings.set_sscha_iterations(params["sscha_iterations"])



"""ALM force constants calculator interface."""

# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import sys
from typing import Optional

import numpy as np

from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive


def run_alm(
    supercell: PhonopyAtoms,
    primitive: Primitive,
    displacements: np.ndarray,
    forces: np.ndarray,
    maxorder: int,
    is_compact_fc: bool = False,
    options: Optional[str] = None,
    log_level: int = 0,
) -> list[np.ndarray]:
    """Calculate force constants using ALM.

    The details of the parameters are found in the ALMFCSolver class.

    """
    if log_level:
        print(
            "---------------------------------"
            " ALM start "
            "--------------------------------"
        )
        print("ALM is a non-trivial force constants calculator. Please cite the paper:")
        print("T. Tadano and S. Tsuneyuki, J. Phys. Soc. Jpn. 87, 041015 (2018).")
        print("ALM is developed at https://github.com/ttadano/ALM by T. Tadano.")

    if log_level == 1:
        print("Increase log-level to watch detailed ALM log.")

    alm = ALMFCSolver(
        supercell,
        primitive,
        displacements,
        forces,
        maxorder,
        is_compact_fc=is_compact_fc,
        options=options,
        log_level=log_level,
    )

    if log_level:
        print(
            "----------------------------------"
            " ALM end "
            "---------------------------------"
        )

    return alm.force_constants


class ALMFCSolver:
    """ALM force constants calculator interface."""

    def __init__(
        self,
        supercell: PhonopyAtoms,
        primitive: Primitive,
        displacements: np.ndarray,
        forces: np.ndarray,
        maxorder: int,
        is_compact_fc: bool = False,
        options: Optional[str] = None,
        log_level: int = 0,
    ):
        """Init method.

        Parameters
        ----------
        supercell : PhonopyAtoms
            Supercell structure.
        primitive : Primitive
            Primitive structure.
        displacements : np.ndarray
            Atomic displacements.
        forces : np.ndarray
            Forces.
        maxorder : int
            Maximum order of force constants. The integer values 1, 2, ...
            correspond to 2nd, 3rd, ... order force constants.
        is_compact_fc : bool, optional
            Compact force constants. Default is False.
        options : str, optional
            ALM options. Default is None.
        log_level : int, optional
            Log level. Default is 0.

        """
        fcs = self._run(
            supercell,
            primitive,
            displacements,
            forces,
            maxorder,
            is_compact_fc=is_compact_fc,
            options=options,
            log_level=log_level,
        )
        self._fc = {i + 2: fc for i, fc in enumerate(fcs)}

    @property
    def force_constants(self) -> dict[int, np.ndarray]:
        """Return force constants.

        Returns
        -------
        dict[int, np.ndarray]
            Force constants with order as key.

        """
        return self._fc

    def _run(
        self,
        supercell: PhonopyAtoms,
        primitive: Primitive,
        displacements: np.ndarray,
        forces: np.ndarray,
        maxorder: int,
        is_compact_fc: bool = False,
        options: Optional[str] = None,
        log_level: int = 0,
    ) -> list[np.ndarray]:
        """Calculate force constants of arbitrary-order using ALM."""
        fcs = None  # This is returned.

        lattice = supercell.cell
        positions = supercell.scaled_positions
        numbers = supercell.numbers
        natom = len(numbers)
        p2s_map = primitive.p2s_map
        p2p_map = primitive.p2p_map

        alm_options = _update_options(options)
        num_elems = len(np.unique(numbers))

        if "norder" in alm_options:
            _maxorder = alm_options["norder"]
        elif "maxorder" in alm_options:
            _maxorder = alm_options["maxorder"]
        else:
            _maxorder = maxorder

        shape = (_maxorder, num_elems, num_elems)
        cutoff_radii = -np.ones(shape, dtype="double")
        if alm_options["cutoff"] is not None:
            if len(alm_options["cutoff"]) == 1:
                cutoff_radii[:] = alm_options["cutoff"][0]
            elif len(alm_options["cutoff"]) == _maxorder:
                for i, cutoff in enumerate(alm_options["cutoff"]):
                    cutoff_radii[i] = cutoff
            elif np.prod(shape) == len(alm_options["cutoff"]):
                cutoff_radii[:] = np.reshape(alm_options["cutoff"], shape)
            else:
                raise RuntimeError("Cutoff is not properly set.")

        _disps, _forces, df_msg = _slice_displacements_and_forces(
            displacements,
            forces,
            alm_options["ndata"],
            alm_options["nstart"],
            alm_options["nend"],
        )

        if log_level > 1:
            print("")
            print("  ndata: %d" % len(displacements))
            for key, val in alm_options.items():
                if val is not None:
                    print("  %s: %s" % (key, val))
            print("")
            print(" " + "-" * 67)

        if log_level > 0:
            log_level_alm = log_level - 1
        else:
            log_level_alm = 0

        try:
            from alm import ALM, optimizer_control_data_types
        except ImportError as exc:
            raise ModuleNotFoundError("ALM python module was not found.") from exc

        with ALM(lattice, positions, numbers) as alm:
            if log_level > 0:
                if alm_options["cutoff"] is not None:
                    for i in range(_maxorder):
                        if _maxorder > 1:
                            print("fc%d" % (i + 2))
                        print(
                            ("cutoff" + " %6s" * num_elems)
                            % tuple(alm.kind_names.values())
                        )
                        for r, kn in zip(cutoff_radii[i], alm.kind_names.values()):
                            print(
                                ("   %-3s" + " %6.2f" * num_elems) % ((kn,) + tuple(r))
                            )
                if df_msg is not None:
                    print(df_msg)
            if log_level > 1:
                print("")
            sys.stdout.flush()

            alm.output_filename_prefix = alm_options["output_filename_prefix"]
            alm.verbosity = log_level_alm

            alm.define(
                _maxorder,
                cutoff_radii=cutoff_radii,
                nbody=alm_options["nbody"],
                symmetrization_basis=alm_options["symmetrization_basis"],
            )
            alm.displacements = _disps
            alm.forces = _forces

            # Mainly for elastic net (or lasso) regression
            optcontrol = {}
            for key in optimizer_control_data_types:
                if key in alm_options:
                    optcontrol[key] = alm_options[key]
            if optcontrol:
                alm.optimizer_control = optcontrol
                if (
                    "cross_validation" in optcontrol
                    and optcontrol["cross_validation"] > 0
                ):
                    alm.optimize(solver=alm_options["solver"])
                    optcontrol["cross_validation"] = 0
                    optcontrol["l1_alpha"] = alm.cv_l1_alpha
                    alm.optimizer_control = optcontrol

            if alm_options["iconst"] == 0:
                alm.set_constraint(translation=False)

            alm.optimize(solver=alm_options["solver"])

            fcs = _extract_fc_from_alm(
                alm, natom, maxorder, is_compact_fc, p2s_map=p2s_map, p2p_map=p2p_map
            )

        return fcs


def _update_options(fc_calculator_options):
    """Set ALM options with appropriate data types.

    fc_calculator_options : str
        This string should be written such as follows:

            "solver = dense, cutoff = 5"

        This string is parsed as collection of settings that are separated by
        comma ','. Each setting has the format of 'option = value'. The value
        is cast to have its appropriate data type for ALM in this method.

    """
    try:
        from alm import optimizer_control_data_types
    except ImportError as exc:
        raise ModuleNotFoundError("ALM python module was not found.") from exc

    # Default settings.
    alm_options = {
        "solver": "dense",
        "ndata": None,
        "nstart": None,
        "nend": None,
        "nbody": None,
        "cutoff": None,
        "symmetrization_basis": "Lattice",
        "output_filename_prefix": None,
        "iconst": 11,
    }

    if fc_calculator_options is not None:
        alm_option_types = {
            "cutoff": np.double,
            "maxorder": int,
            "norder": int,
            "ndata": int,
            "nstart": int,
            "nend": int,
            "nbody": np.dtype("int64"),
            "output_filename_prefix": str,
            "solver": str,
            "symmetrization_basis": str,
            "iconst": int,
        }
        alm_option_types.update(optimizer_control_data_types)
        for option_str in fc_calculator_options.split(","):
            key, val = [x.strip() for x in option_str.split("=")[:2]]
            if key.lower() in alm_option_types:
                if alm_option_types[key.lower()] is np.double:
                    option_value = np.array(
                        [float(x) for x in val.split()], dtype="double"
                    )
                elif alm_option_types[key.lower()] is np.dtype("int64"):
                    option_value = np.array(
                        [int(x) for x in val.split()], dtype="int64"
                    )
                else:
                    option_value = alm_option_types[key.lower()](val)
                alm_options[key] = option_value
    return alm_options


def _slice_displacements_and_forces(d, f, ndata, nstart, nend):
    msg = None
    if ndata is not None:
        _d = d[:ndata]
        _f = f[:ndata]
        msg = "Number of displacement supercells: %d" % ndata
    elif nstart is not None and nend is not None:
        _d = d[nstart - 1 : nend]
        _f = f[nstart - 1 : nend]
        msg = "Supercell index range: %d - %d" % (nstart, nend)
    else:
        return d, f, None

    return (
        np.array(_d, dtype="double", order="C"),
        np.array(_f, dtype="double", order="C"),
        msg,
    )


def _extract_fc_from_alm(
    alm, natom, maxorder, is_compact_fc, p2s_map=None, p2p_map=None
) -> list[np.ndarray]:
    # Harmonic: order=1, 3rd: order=2, ...
    fcs = []
    for order in range(1, maxorder + 1):
        fc = None
        p2s_map_alm = alm.getmap_primitive_to_supercell()[0]
        if (
            is_compact_fc
            and len(p2s_map_alm) == len(p2s_map)
            and (p2s_map_alm == p2s_map).all()
        ):
            fc_shape = (len(p2s_map),) + (natom,) * order + (3,) * (order + 1)
            fc = np.zeros(fc_shape, dtype="double", order="C")
            for fc_elem, indices in zip(
                *alm.get_fc(order, mode="origin", permutation=1)
            ):
                v = indices // 3
                c = indices % 3
                selection = (p2p_map[v[0]],) + tuple(v[1:]) + tuple(c)
                fc[selection] = fc_elem

        if fc is None:
            if is_compact_fc:
                atom_list = p2s_map
            else:
                atom_list = np.arange(natom, dtype=int)
            fc_shape = (len(atom_list),) + (natom,) * order + (3,) * (order + 1)
            fc = np.zeros(fc_shape, dtype="double", order="C")
            for fc_elem, indices in zip(*alm.get_fc(order, mode="all", permutation=1)):
                v = indices // 3
                idx = np.where(atom_list == v[0])[0]
                if len(idx) > 0:
                    c = indices % 3
                    selection = (idx[0],) + tuple(v[1:]) + tuple(c)
                    fc[selection] = fc_elem

        fcs.append(fc)

    return fcs



"""Tests of Abinit calculator interface."""

import io
import pathlib

import numpy as np

from phonopy.interface.abinit import get_abinit_structure, read_abinit
from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.structure.cells import isclose

cwd = pathlib.Path(__file__).parent


def test_read_abinit():
    """Test of read_abinit."""
    cell = read_abinit(cwd / "NaCl-abinit.in")
    filename = cwd / "NaCl-abinit-pwscf.yaml"
    cell_ref = read_cell_yaml(filename)
    assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r


def test_get_abinit_structure():
    """Test get_abinit_structure."""
    cell_ref = read_abinit(cwd / "NaCl-abinit.in")
    cell = read_abinit(io.StringIO(get_abinit_structure(cell_ref)))
    isclose(cell_ref, cell)



"""Tests for thermal property calculation."""

import numpy as np

temps = [
    0.000000,
    100.000000,
    200.000000,
    300.000000,
    400.000000,
    500.000000,
    600.000000,
    700.000000,
    800.000000,
    900.000000,
]
fes = [
    4.856522,
    3.916257,
    -0.275662,
    -6.808753,
    -14.961276,
    -24.341220,
    -34.707527,
    -45.897738,
    -57.795132,
    -70.311860,
]
entropies = [
    0.000000,
    26.327525,
    55.256537,
    74.268068,
    88.155267,
    99.052543,
    108.007856,
    115.604464,
    122.198503,
    128.022838,
]
cvs = [
    0.000000,
    36.207244,
    45.673361,
    47.838756,
    48.634184,
    49.009290,
    49.214930,
    49.339570,
    49.420727,
    49.476488,
]


def test_thermal_properties(ph_nacl):
    """Test thermal property calculation with t_step and t_max parameters."""
    ph_nacl.run_mesh([5, 5, 5])
    ph_nacl.run_thermal_properties(t_step=100, t_max=900, cutoff_frequency=1e-5)
    _test_thermal_properties(ph_nacl)


def test_thermal_properties_at_temperatues(ph_nacl):
    """Test thermal property calculation with temperatures parameter."""
    ph_nacl.run_mesh([5, 5, 5])
    temperatures = [0, 100, 200, 300, 400, 500, 600, 700, 800, 900]
    ph_nacl.run_thermal_properties(temperatures=temperatures, cutoff_frequency=1e-5)
    _test_thermal_properties(ph_nacl)


def _test_thermal_properties(ph):
    tp = ph.thermal_properties

    # for vals in tp.thermal_properties:
    #     print(", ".join(["%.6f" % v for v in vals]))

    for i in range(2):
        if i == 1:
            tp.run(lang="py")
        for vals_ref, vals in zip((temps, fes, entropies, cvs), tp.thermal_properties):
            np.testing.assert_allclose(vals_ref, vals, atol=1e-5)



"""Tests for CRYSTAL calculator interface."""

import os

import numpy as np

from phonopy.interface.crystal import read_crystal
from phonopy.interface.phonopy_yaml import read_cell_yaml

data_dir = os.path.dirname(os.path.abspath(__file__))


def test_read_crystal():
    """Test of read_crystal."""
    cell, pp_filenames = read_crystal(os.path.join(data_dir, "Si-CRYSTAL.o"))
    filename = os.path.join(data_dir, "Si-CRYSTAL.yaml")
    cell_ref = read_cell_yaml(filename)
    assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r



"""API for QHA calculation."""

# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

from phonopy.qha.core import QHA, BulkModulus


class PhonopyQHA:
    """PhonopyQHA API."""

    def __init__(
        self,
        volumes=None,
        electronic_energies=None,
        temperatures=None,
        free_energy=None,
        cv=None,
        entropy=None,
        pressure=None,
        eos="vinet",
        t_max=None,
        energy_plot_factor=None,
        verbose=False,
    ):
        """Init method.

        Notes
        -----
        The first two parameters have to be in this order for the backward
        compatibility.

        Parameters
        ----------
        volumes: array_like
            Unit cell volumes (V) in angstrom^3.
            dtype='double'
            shape=(volumes,)
        electronic_energies: array_like
            Electronic energies (U_el) or electronic free energies (F_el) in eV.
            It is assumed as formar if ndim==1 and latter if ndim==2.
            dtype='double'
            shape=(volumes,) or (temperatuers, volumes)
        temperatures: array_like
            Temperatures ascending order (T) in K.
            dtype='double'
            shape=(temperatures,)
        pressure: float,
            Pressure in GPa that is added to energy as PV term.
        free_energy: array_like
            Phonon Helmholtz free energy (F_ph) in kJ/mol.
            dtype='double'
            shape=(temperatuers, volumes)
        cv: array_like
            Phonon heat capacity at constant volume in J/K/mol.
            dtype='double'
            shape=(temperatuers, volumes)
        entropy: array_like
            Phonon entropy at constant volume (S_ph) in J/K/mol.
            dtype='double'
            shape=(temperatuers, volumes)
        pressure: float, optional
            Pressure in GPa that is added to energy as PV term.
        eos: str, optional
            Equation of state used for fitting F vs V.
            'vinet', 'murnaghan' or 'birch_murnaghan'.
        t_max: float
            Maximum temperature to be calculated. This has to be not
            greater than the temperature of the third element from the
            end of 'temperatre' elements. If max_t=None, the temperature
            of the third element from the end is used.
        energy_plot_factor: float
            This value is multiplied to energy like values only in plotting.
        verbose: boolean
            Show log or not.

        """
        self._bulk_modulus = BulkModulus(
            volumes, electronic_energies, pressure=pressure, eos=eos
        )

        if temperatures is not None:
            self._qha = QHA(
                volumes,
                electronic_energies,
                temperatures,
                cv,
                entropy,
                free_energy,
                pressure=pressure,
                eos=eos,
                t_max=t_max,
                energy_plot_factor=energy_plot_factor,
            )
            self._qha.run(verbose=verbose)

    @property
    def bulk_modulus(self):
        """Return bulk modulus computed without phonon contribution.

        Returns
        -------
        float
            Bulk modulus calculated without phonon contribution.

        """
        return self._bulk_modulus.bulk_modulus

    @property
    def thermal_expansion(self):
        """Return thermal expansion coefficients at temperatures.

        Returns
        -------
        list
            Thermal expansion coefficients at temperatues.
            shape=(temperatures, )

        """
        return self._qha.thermal_expansion

    @property
    def helmholtz_volume(self):
        """Return free_energies at volumes.

        Returns
        -------
        ndarray
            Helmholtz free energies calculated at temperatures and volumes.
            shape=(temperatures, volumes)

        """
        return self._qha.helmholtz_volume

    @property
    def volume_temperature(self):
        """Return volumes at temperatures.

        Returns
        -------
        ndarray
            Equilibrium volumes at temperatures
            shape=(temperatures,), dtype=float

        """
        return self._qha.volume_temperature

    @property
    def gibbs_temperature(self):
        """Return Gibbs free energies at temperatures.

        Returns
        -------
        ndarray
            Gibbs free energies at temperatures.
            shape=(temperatures, ), dtype=float

        """
        return self._qha.gibbs_temperature

    @property
    def bulk_modulus_temperature(self):
        """Return bulk modulus at temperatures.

        Returns
        -------
        ndarray
            Bulk modulus at constant pressure and temperatures
            shape=(temperatures, ), dtype=float

        """
        return self._qha.bulk_modulus_temperature

    @property
    def heat_capacity_P_numerical(self):
        """Return heat capacities at constant pressure at temperatures.

        These values are calculated by -T*d^2G/dT^2.

        Returns
        -------
        list
            Heat capacity at constant pressure and temperatures.
            shape=(temperatures, )

        """
        return self._qha.get_heat_capacity_P_numerical()

    @property
    def heat_capacity_P_polyfit(self):
        """Return heat capacities at constant pressure at temperatures.

        Note
        ----
        This does not work when temperature dependent electronic_energies
        is supplied.

        Returns
        -------
        list
            Heat capacities at constant pressure at temperatures, which are
            calculated from the values obtained by polynomial fittings of
            Cv and S.

            shape=(temperatures, )

        """
        return self._qha.heat_capacity_P_polyfit

    @property
    def gruneisen_temperature(self):
        """Return Gruneisen parameters at temperatures.

        Returns
        -------
        list
            Thermodynamic Gruneisen parameters at temperatures.
            shape=(temperatures, )

        """
        return self._qha.gruneisen_temperature

    def get_bulk_modulus_parameters(self):
        """Return temperature independent bulk modulus EOS fitting parameters.

        These values are those computed without phonon free energy.

        (lowest energy, bulk modulus, b_prime, equilibrium volume)

        """
        return self._bulk_modulus.get_parameters()

    def write_helmholtz_volume(self, filename="helmholtz-volume.dat"):
        """Write Helmholtz free energy vs volume in file."""
        self._qha.write_helmholtz_volume(filename=filename)

    def write_helmholtz_volume_fitted(
        self, thin_number, filename="helmholtz-volume_fitted.dat"
    ):
        """Write Helmholtz free energy (fitted) vs volume in file."""
        self._qha.write_helmholtz_volume_fitted(thin_number, filename=filename)

    def write_volume_temperature(self, filename="volume-temperature.dat"):
        """Write volume vs temperature in file."""
        self._qha.write_volume_temperature(filename=filename)

    def write_thermal_expansion(self, filename="thermal_expansion.dat"):
        """Write thermal expansion vs temperature in file."""
        self._qha.write_thermal_expansion(filename=filename)

    def write_gibbs_temperature(self, filename="gibbs-temperature.dat"):
        """Write Gibbs free energy vs temperature in file."""
        self._qha.write_gibbs_temperature(filename=filename)

    def write_bulk_modulus_temperature(self, filename="bulk_modulus-temperature.dat"):
        """Write bulk modulus vs temperature in file."""
        self._qha.write_bulk_modulus_temperature(filename=filename)

    def plot_bulk_modulus(self, thin_number=10):
        """Return pyplot of bulk modulus fitting curve."""
        return self._bulk_modulus.plot(thin_number=thin_number)

    def plot_qha(self, thin_number=10, volume_temp_exp=None):
        """Return pyplot of QHA fitting curves at temperatures."""
        return self._qha.plot(thin_number=thin_number, volume_temp_exp=volume_temp_exp)

    def plot_helmholtz_volume(
        self, thin_number=10, xlabel=r"Volume $(\AA^3)$", ylabel="Free energy"
    ):
        """Return pyplot of Helmholtz free energes vs volume at temperatures."""
        return self._qha.plot_helmholtz_volume(
            thin_number=thin_number, xlabel=xlabel, ylabel=ylabel
        )

    def plot_pdf_helmholtz_volume(
        self, thin_number=10, filename="helmholtz-volume.pdf"
    ):
        """Plot Helmholtz free energes vs volume at temperatures in pdf."""
        self._qha.plot_pdf_helmholtz_volume(thin_number=thin_number, filename=filename)

    def plot_volume_temperature(self, exp_data=None):
        """Return pyplot of volume vs temperature."""
        return self._qha.plot_volume_temperature(exp_data=exp_data)

    def plot_pdf_volume_temperature(
        self, exp_data=None, filename="volume-temperature.pdf"
    ):
        """Plot volume vs temperature in pdf."""
        self._qha.plot_pdf_volume_temperature(exp_data=exp_data, filename=filename)

    def plot_thermal_expansion(self):
        """Return pyplot of thermal expansion vs temperature."""
        return self._qha.plot_thermal_expansion()

    def plot_pdf_thermal_expansion(self, filename="thermal_expansion.pdf"):
        """Plot thermal expansion vs temperature in pdf."""
        self._qha.plot_pdf_thermal_expansion(filename=filename)

    def plot_gibbs_temperature(
        self, xlabel="Temperature (K)", ylabel="Gibbs free energy"
    ):
        """Return pyplot of Gibbs free energy vs temperature."""
        return self._qha.plot_gibbs_temperature(xlabel=xlabel, ylabel=ylabel)

    def plot_pdf_gibbs_temperature(self, filename="gibbs-temperature.pdf"):
        """Plot Gibbs free energy vs temperature in pdf."""
        self._qha.plot_pdf_gibbs_temperature(filename=filename)

    def plot_bulk_modulus_temperature(
        self, xlabel="Temperature (K)", ylabel="Bulk modulus"
    ):
        """Return pyplot of bulk modulus vs temperature."""
        return self._qha.plot_bulk_modulus_temperature(xlabel=xlabel, ylabel=ylabel)

    def plot_pdf_bulk_modulus_temperature(
        self, filename="bulk_modulus-temperature.pdf"
    ):
        """Plot bulk modulus vs temperature in pdf."""
        self._qha.plot_pdf_bulk_modulus_temperature(filename=filename)

    def plot_heat_capacity_P_numerical(self, Z=1, exp_data=None):
        """Return pyplot of C_P by numerical difference vs temperature."""
        return self._qha.plot_heat_capacity_P_numerical(Z=Z, exp_data=exp_data)

    def plot_pdf_heat_capacity_P_numerical(
        self, exp_data=None, filename="Cp-temperature.pdf"
    ):
        """Plot C_P by numerical difference vs temperature in pdf."""
        self._qha.plot_pdf_heat_capacity_P_numerical(
            exp_data=exp_data, filename=filename
        )

    def write_heat_capacity_P_numerical(self, filename="Cp-temperature.dat"):
        """Write C_P by numerical difference vs temperature in file."""
        self._qha.write_heat_capacity_P_numerical(filename=filename)

    def plot_heat_capacity_P_polyfit(self, exp_data=None, Z=1):
        """Return pyplot of C_P by fittings vs temperature."""
        return self._qha.plot_heat_capacity_P_polyfit(Z=Z, exp_data=exp_data)

    def plot_pdf_heat_capacity_P_polyfit(
        self, exp_data=None, filename="Cp-temperature_polyfit.pdf"
    ):
        """Plot C_P by fittings vs temperature in pdf."""
        self._qha.plot_pdf_heat_capacity_P_polyfit(exp_data=exp_data, filename=filename)

    def write_heat_capacity_P_polyfit(
        self,
        filename="Cp-temperature_polyfit.dat",
        filename_ev="entropy-volume.dat",
        filename_cvv="Cv-volume.dat",
        filename_dsdvt="dsdv-temperature.dat",
    ):
        """Write C_P by fittings vs temperature in file."""
        self._qha.write_heat_capacity_P_polyfit(
            filename=filename,
            filename_ev=filename_ev,
            filename_cvv=filename_cvv,
            filename_dsdvt=filename_dsdvt,
        )

    def plot_gruneisen_temperature(self):
        """Return pyplot of Grueneisen parameter vs temperature."""
        return self._qha.plot_gruneisen_temperature()

    def plot_pdf_gruneisen_temperature(self, filename="gruneisen-temperature.pdf"):
        """Plot Grueneisen parameter vs temperature in pdf."""
        self._qha.plot_pdf_gruneisen_temperature(filename=filename)

    def write_gruneisen_temperature(self, filename="gruneisen-temperature.dat"):
        """Write Grueneisen parameter vs temperature in file."""
        self._qha.write_gruneisen_temperature(filename=filename)

    def get_bulk_modulus(self):
        """Return bulk moduli with no phonon contribution."""
        warnings.warn(
            "PhonopyQHA.get_bulk_modulus() is deprecated.Use bulk_modulus attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.bulk_modulus

    def get_helmholtz_volume(self):
        """Return Helmholtz free energies at temperatures and volumes."""
        warnings.warn(
            "PhonopyQHA.get_helmholtz_volume() is deprecated."
            "Use helmholtz_volume attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.helmholtz_volume

    def get_volume_temperature(self):
        """Return equilibrium volumes at temperatures."""
        warnings.warn(
            "PhonopyQHA.get_volume_temperature() is deprecated."
            "Use volume_temperature attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.volume_temperature

    def get_thermal_expansion(self):
        """Return thermal expansion coefficients at temperatures."""
        warnings.warn(
            "PhonopyQHA.get_thermal_expansion() is deprecated."
            "Use thermal_expansion attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.thermal_expansion

    def get_gibbs_temperature(self):
        """Return Gibbs free energies at temperatures."""
        warnings.warn(
            "PhonopyQHA.get_gibbs_temperature() is deprecated."
            "Use gibbs_temperature attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.gibbs_temperature

    def get_bulk_modulus_temperature(self):
        """Return bulk moduli at temperatures."""
        warnings.warn(
            "PhonopyQHA.get_bulk_modulus_temperature() is deprecated."
            "Use bulk_modulus_temperature attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.bulk_modulus_temperature

    def get_heat_capacity_P_numerical(self):
        """Return C_P calculated by numerical differentiation."""
        warnings.warn(
            "PhonopyQHA.get_heat_capacity_P_numerical() is deprecated."
            "Use heat_capacity_P_numerical attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.heat_capacity_P_numerical

    def get_heat_capacity_P_polyfit(self):
        """Return C_P calculated by fittings."""
        warnings.warn(
            "PhonopyQHA.get_heat_capacity_P_polyfit() is deprecated."
            "Use heat_capacity_P_polyfit attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.heat_capacity_P_polyfit

    def get_gruneisen_temperature(self):
        """Return Grueneisen paramters at temperatures."""
        warnings.warn(
            "PhonopyQHA.get_gruneisen_temperature() is deprecated."
            "Use gruneisen_temperature attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.gruneisen_temperature



"""Tests of routines in cells.py."""

import os
from collections.abc import Callable

import numpy as np
import pytest

from phonopy import Phonopy
from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import (
    Primitive,
    ShortestPairs,
    TrimmedCell,
    compute_all_sg_permutations,
    compute_permutation_for_rotation,
    convert_to_phonopy_primitive,
    dense_to_sparse_svecs,
    get_angles,
    get_cell_matrix_from_lattice,
    get_cell_parameters,
    get_primitive,
    get_primitive_matrix,
    get_supercell,
    isclose,
    sparse_to_dense_svecs,
)

data_dir = os.path.dirname(os.path.abspath(__file__))
primitive_matrix_nacl = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]

multi_nacl_ref = [
    1,
    1,
    2,
    1,
    2,
    1,
    4,
    1,
    2,
    1,
    4,
    1,
    4,
    1,
    8,
    1,
    1,
    1,
    2,
    1,
    1,
    2,
    2,
    2,
    1,
    2,
    2,
    2,
    1,
    4,
    2,
    4,
    1,
    1,
    1,
    2,
    2,
    1,
    2,
    2,
    1,
    2,
    1,
    4,
    2,
    2,
    2,
    4,
    1,
    1,
    1,
    2,
    1,
    2,
    1,
    4,
    2,
    1,
    2,
    2,
    2,
    2,
    2,
    4,
    1,
    1,
    1,
    2,
    1,
    2,
    1,
    4,
    1,
    2,
    1,
    4,
    1,
    4,
    1,
    8,
    1,
    1,
    1,
    2,
    2,
    1,
    2,
    2,
    2,
    1,
    2,
    2,
    4,
    1,
    4,
    2,
    1,
    1,
    2,
    1,
    1,
    2,
    2,
    2,
    2,
    1,
    4,
    1,
    2,
    2,
    4,
    2,
    1,
    1,
    2,
    1,
    2,
    1,
    4,
    1,
    1,
    2,
    2,
    2,
    2,
    2,
    4,
    2,
]
svecs_nacl_ref10 = [[-0.5, 0.0, 0.0], [0.5, 0.0, 0.0]]
svecs_nacl_ref30 = [
    [-0.5, -0.5, 0.0],
    [-0.5, 0.5, 0.0],
    [0.5, -0.5, 0.0],
    [0.5, 0.5, 0.0],
]


def test_compute_permutation_sno2(ph_sno2: Phonopy):
    """Test of compute_permutation by SnO2."""
    _test_compute_permutation(ph_sno2)


def test_compute_permutation_tio2(ph_tio2: Phonopy):
    """Test of compute_permutation by TiO2."""
    _test_compute_permutation(ph_tio2)


def test_compute_permutation_nacl(ph_nacl: Phonopy):
    """Test of compute_permutation by NaCl."""
    _test_compute_permutation(ph_nacl)


def _test_compute_permutation(ph: Phonopy):
    symmetry = ph.primitive_symmetry
    ppos = ph.primitive.scaled_positions
    plat = ph.primitive.cell.T
    symprec = symmetry.tolerance
    rots = symmetry.symmetry_operations["rotations"]
    trans = symmetry.symmetry_operations["translations"]
    perms = compute_all_sg_permutations(ppos, rots, trans, plat, symprec)
    for i, (r, t) in enumerate(zip(rots, trans)):
        ppos_rot = np.dot(ppos, r.T) + t
        perm = compute_permutation_for_rotation(ppos, ppos_rot, plat, symprec)
        np.testing.assert_array_equal(perms[i], perm)
        diff = ppos[perm] - ppos_rot
        diff -= np.rint(diff)
        assert ((np.dot(diff, plat) ** 2).sum(axis=1) < symprec).all()


@pytest.mark.parametrize("nosnf", [True, False])
def test_get_supercell_convcell_sio2(
    convcell_sio2: PhonopyAtoms, nosnf, helper_methods
):
    """Test of get_supercell with/without SNF by SiO2."""
    _test_get_supercell_convcell_sio2(convcell_sio2, helper_methods, is_old_style=nosnf)


@pytest.mark.parametrize("nosnf", [True, False])
def test_get_supercell_primcell_si(
    primcell_si: PhonopyAtoms, nosnf, helper_methods: Callable
):
    """Test of get_supercell with/without SNF by Si."""
    _test_get_supercell_primcell_si(primcell_si, helper_methods, is_old_style=nosnf)


def test_get_supercell_nacl_snf(
    nacl_unitcell_order1: PhonopyAtoms, helper_methods: Callable
):
    """Test of get_supercell using SNF by NaCl."""
    cell = nacl_unitcell_order1
    smat = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
    scell = get_supercell(cell, smat, is_old_style=True)
    scell_snf = get_supercell(cell, smat, is_old_style=False)
    helper_methods.compare_cells(scell, scell_snf)


@pytest.mark.parametrize("is_ncl", [False, True])
def test_get_supercell_Cr_with_magmoms(
    convcell_cr: PhonopyAtoms, is_ncl: bool, helper_methods: Callable
):
    """Test of get_supercell using SNF by Cr with magnetic moments."""
    if is_ncl:
        convcell_cr.magnetic_moments = [[0, 0, 1], [0, 0, -1]]
        ref_magmoms = [[0, 0, 1]] * 4 + [[0, 0, -1]] * 4
    else:
        convcell_cr.magnetic_moments = [1, -1]
        ref_magmoms = [1.0] * 4 + [-1.0] * 4

    smat = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
    scell = get_supercell(convcell_cr, smat, is_old_style=True)

    np.testing.assert_allclose(
        scell.magnetic_moments,
        ref_magmoms,
        atol=1e-8,
    )
    scell_snf = get_supercell(convcell_cr, smat, is_old_style=False)
    helper_methods.compare_cells(scell, scell_snf)
    convcell_cr.magnetic_moments = None


def _test_get_supercell_convcell_sio2(
    convcell_sio2: PhonopyAtoms, helper_methods, is_old_style=True
):
    smat = np.diag([1, 2, 3])
    fname = "SiO2-123.yaml"
    scell = get_supercell(convcell_sio2, smat, is_old_style=is_old_style)
    cell_ref = read_cell_yaml(os.path.join(data_dir, fname))
    if is_old_style is True:
        helper_methods.compare_cells_with_order(scell, cell_ref)
    else:
        helper_methods.compare_cells(scell, cell_ref)


def _test_get_supercell_primcell_si(
    primcell_si: PhonopyAtoms, helper_methods, is_old_style=True
):
    smat = [[-1, 1, 1], [1, -1, 1], [1, 1, -1]]
    fname = "Si-conv.yaml"
    scell = get_supercell(primcell_si, smat, is_old_style=is_old_style)
    cell_ref = read_cell_yaml(os.path.join(data_dir, fname))
    if is_old_style is True:
        helper_methods.compare_cells_with_order(scell, cell_ref)
    else:
        helper_methods.compare_cells(scell, cell_ref)


def test_get_primitive_convcell_nacl(
    nacl_unitcell_order1: PhonopyAtoms, primcell_nacl: PhonopyAtoms, helper_methods
):
    """Test get_primitive by NaCl."""
    pcell = get_primitive(nacl_unitcell_order1, primitive_matrix=primitive_matrix_nacl)
    helper_methods.compare_cells_with_order(pcell, primcell_nacl)


def test_get_primitive_convcell_nacl_with_cetring_symbol(
    nacl_unitcell_order1: PhonopyAtoms, primcell_nacl: PhonopyAtoms, helper_methods
):
    """Test get_primitive by NaCl."""
    pcell = get_primitive(nacl_unitcell_order1, primitive_matrix="F")
    helper_methods.compare_cells_with_order(pcell, primcell_nacl)


@pytest.mark.parametrize("is_ncl", [False, True])
def test_get_primitive_convcell_Cr_with_magmoms(
    convcell_cr: PhonopyAtoms, is_ncl: bool, helper_methods: Callable
):
    """Test get_primitive by Cr with magmoms."""
    if is_ncl:
        convcell_cr.magnetic_moments = [[0, 0, 1], [0, 0, -1]]
    else:
        convcell_cr.magnetic_moments = [1, -1]
    smat = [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    scell = get_supercell(convcell_cr, smat, is_old_style=True)
    pmat = np.linalg.inv(smat)
    pcell = get_primitive(scell, primitive_matrix=pmat)
    helper_methods.compare_cells(convcell_cr, pcell)
    convcell_cr.magnetic_moments = None


@pytest.mark.parametrize("store_dense_svecs", [True, False])
def test_get_primitive_convcell_nacl_svecs(
    nacl_unitcell_order1: PhonopyAtoms, store_dense_svecs
):
    """Test shortest vectors by NaCl."""
    pcell = get_primitive(
        nacl_unitcell_order1,
        primitive_matrix=primitive_matrix_nacl,
        store_dense_svecs=store_dense_svecs,
    )
    svecs, multi = pcell.get_smallest_vectors()
    if store_dense_svecs:
        assert svecs.shape == (54, 3)
        assert multi.shape == (8, 2, 2)
        assert np.sum(multi[:, :, 0]) == 54
        assert np.sum(multi[-1:, -1, :]) == 54
    else:
        assert svecs.shape == (8, 2, 27, 3)
        assert multi.shape == (8, 2)


def test_TrimmedCell(nacl_unitcell_order1: PhonopyAtoms, helper_methods: Callable):
    """Test TrimmedCell by NaCl."""
    pmat = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    smat2 = np.eye(3, dtype="intc") * 2
    pmat2 = np.dot(np.linalg.inv(smat2), pmat)
    smat3 = np.eye(3, dtype="intc") * 3
    pmat3 = np.dot(np.linalg.inv(smat3), pmat)

    cell = nacl_unitcell_order1
    scell2 = get_supercell(cell, smat2)
    scell3 = get_supercell(cell, smat3)
    n = len(scell3) // 2
    # swap first and last half of atomic order
    indices = [i + n for i in range(n)] + list(range(n))
    scell3_swap = PhonopyAtoms(
        cell=scell3.cell,
        scaled_positions=scell3.scaled_positions[indices],
        symbols=[scell3.symbols[i] for i in indices],
    )
    tcell2 = TrimmedCell(pmat2, scell2)
    tcell3 = TrimmedCell(
        pmat3, scell3_swap, positions_to_reorder=tcell2.scaled_positions
    )
    helper_methods.compare_cells_with_order(tcell2, tcell3)


def test_ShortestPairs_sparse_nacl(ph_nacl: Phonopy, helper_methods: Callable):
    """Test ShortestPairs (parse) by NaCl."""
    scell = ph_nacl.supercell
    pcell = ph_nacl.primitive
    pos = scell.scaled_positions
    spairs = ShortestPairs(scell.cell, pos, pos[pcell.p2s_map], store_dense_svecs=False)
    svecs = spairs.shortest_vectors
    multi = spairs.multiplicities
    np.testing.assert_array_equal(multi.ravel(), multi_nacl_ref)
    pos_from_svecs = svecs[:, 0, 0, :] + pos[0]
    np.testing.assert_allclose(svecs_nacl_ref10, svecs[1, 0, :2], atol=1e-8)
    np.testing.assert_allclose(svecs_nacl_ref30, svecs[3, 0, :4], atol=1e-8)
    helper_methods.compare_positions_with_order(pos_from_svecs, pos, scell.cell)


def test_ShortestPairs_dense_nacl(ph_nacl: Phonopy, helper_methods: Callable):
    """Test ShortestPairs (dense) by NaCl."""
    scell = ph_nacl.supercell
    pcell = ph_nacl.primitive
    pos = scell.scaled_positions
    spairs = ShortestPairs(scell.cell, pos, pos[pcell.p2s_map], store_dense_svecs=True)
    svecs = spairs.shortest_vectors
    multi = spairs.multiplicities
    assert multi[-1, -1, :].sum() == multi[:, :, 0].sum()
    np.testing.assert_array_equal(multi[:, :, 0].ravel(), multi_nacl_ref)
    np.testing.assert_allclose(
        svecs_nacl_ref10, svecs[multi[1, 0, 1] : multi[1, 0, :].sum()], atol=1e-8
    )
    np.testing.assert_allclose(
        svecs_nacl_ref30, svecs[multi[3, 0, 1] : multi[3, 0, :].sum()], atol=1e-8
    )
    pos_from_svecs = svecs[multi[:, 0, 1], :] + pos[0]
    helper_methods.compare_positions_with_order(pos_from_svecs, pos, scell.cell)


def test_sparse_to_dense_and_dense_to_sparse_nacl(ph_nacl: Phonopy):
    """Test for sparse_to_dense_svecs and dense_to_sparse_svecs by NaCl."""
    _test_sparse_to_dense_and_dense_to_sparse(ph_nacl.supercell, ph_nacl.primitive)


def test_sparse_to_dense_and_dense_to_sparse_tipn3(ph_tipn3: Phonopy):
    """Test for sparse_to_dense_svecs and dense_to_sparse_svecs by TiPN3."""
    _test_sparse_to_dense_and_dense_to_sparse(ph_tipn3.supercell, ph_tipn3.primitive)


def test_sparse_to_dense_and_dense_to_sparse_al2o3(convcell_al2o3: PhonopyAtoms):
    """Test for sparse_to_dense_svecs and dense_to_sparse_svecs by Al2O3."""
    smat = np.diag([3, 3, 2])
    scell = get_supercell(convcell_al2o3, smat)

    pmat = np.diag([1.0 / 3, 1.0 / 3, 1.0 / 2]) @ get_primitive_matrix("R")
    pcell = get_primitive(scell, primitive_matrix=pmat)
    _test_sparse_to_dense_and_dense_to_sparse(scell, pcell)

    pmat = np.diag([1.0 / 3, 1.0 / 3, 1.0 / 2])
    pcell = get_primitive(scell, primitive_matrix=pmat)
    _test_sparse_to_dense_and_dense_to_sparse(scell, pcell)


def _test_sparse_to_dense_and_dense_to_sparse(scell: PhonopyAtoms, pcell: Primitive):
    pos = scell.scaled_positions

    spairs = ShortestPairs(scell.cell, pos, pos[pcell.p2s_map], store_dense_svecs=False)
    svecs = spairs.shortest_vectors
    multi = spairs.multiplicities

    spairs = ShortestPairs(scell.cell, pos, pos[pcell.p2s_map], store_dense_svecs=True)
    dsvecs = spairs.shortest_vectors
    dmulti = spairs.multiplicities

    _dsvecs, _dmulti = sparse_to_dense_svecs(svecs, multi)

    np.testing.assert_array_equal(dmulti, _dmulti)
    np.testing.assert_allclose(dsvecs, _dsvecs, rtol=0, atol=1e-8)

    _ssvecs, _smulti = dense_to_sparse_svecs(dsvecs, dmulti)

    np.testing.assert_array_equal(multi, _smulti)
    np.testing.assert_allclose(svecs, _ssvecs, rtol=0, atol=1e-8)


def test_isclose(ph_nacl: Phonopy):
    """Test of isclose wit same order of atoms.."""
    scell = ph_nacl.supercell
    pcell = ph_nacl.primitive
    assert isclose(pcell, pcell)
    assert isclose(scell, scell)
    assert not isclose(scell, pcell)


def test_isclose_with_arbitrary_order(
    nacl_unitcell_order1: PhonopyAtoms, nacl_unitcell_order2: PhonopyAtoms
):
    """Test of isclose with different order."""
    cell1 = nacl_unitcell_order1
    cell2 = nacl_unitcell_order2
    assert not isclose(cell1, cell2)
    _isclose = isclose(cell1, cell2, with_arbitrary_order=True)
    assert isinstance(_isclose, bool)
    assert _isclose
    order = isclose(cell1, cell2, with_arbitrary_order=True, return_order=True)
    np.testing.assert_array_equal(order, [0, 4, 1, 5, 2, 6, 3, 7])


def test_convert_to_phonopy_primitive(ph_nacl: Phonopy):
    """Test for convert_to_phonopy_primitive."""
    scell = ph_nacl.supercell
    pcell = ph_nacl.primitive
    _pcell = convert_to_phonopy_primitive(scell, pcell)
    assert isclose(pcell, _pcell)

    # Changing order of atoms is not allowed.
    points = pcell.scaled_positions[[1, 0]]
    symbols = [pcell.symbols[i] for i in (1, 0)]
    cell = pcell.cell
    pcell_mode = PhonopyAtoms(cell=cell, scaled_positions=points, symbols=symbols)
    with pytest.raises(RuntimeError):
        _pcell = convert_to_phonopy_primitive(scell, pcell_mode)


def test_get_cell_matrix_from_lattice(primcell_nacl: PhonopyAtoms):
    """Test for test_get_cell_matrix_from_lattice."""
    pcell = primcell_nacl
    lattice = get_cell_matrix_from_lattice(pcell.cell)
    np.testing.assert_allclose(
        get_angles(lattice, is_radian=False),
        get_angles(pcell.cell, is_radian=False),
        atol=1e-8,
    )
    np.testing.assert_allclose(
        get_angles(lattice, is_radian=True),
        get_angles(pcell.cell, is_radian=True),
        atol=1e-8,
    )
    np.testing.assert_allclose(
        get_cell_parameters(lattice), get_cell_parameters(pcell.cell), atol=1e-8
    )
    np.testing.assert_allclose(
        [
            [4.02365076, 0.0, 0.0],
            [2.01182538, 3.48458377, 0.0],
            [2.01182538, 1.16152792, 3.28529709],
        ],
        lattice,
        atol=1e-7,
    )


def test_get_supercell_with_Xn_symbol(ph_nacl: Phonopy):
    """Test of get_supercell with Xn symbol."""
    symbols = ph_nacl.unitcell.symbols
    symbols[-1] = "Cl1"
    masses = ph_nacl.unitcell.masses
    masses[-1] = 70.0
    cell = PhonopyAtoms(
        cell=ph_nacl.unitcell.cell,
        scaled_positions=ph_nacl.unitcell.scaled_positions,
        symbols=symbols,
        masses=masses,
    )
    scell = get_supercell(cell, np.diag([2, 2, 2]))
    assert scell.symbols[-8:] == ["Cl1"] * 8
    np.testing.assert_allclose(scell.masses[-8:], [70.0] * 8)


def test_get_primitive_with_Xn_symbol(ph_nacl: Phonopy):
    """Test of get_primitive with Xn symbol.

    Symbols with index breaks symmetry to make primitive cell.

    Can not make primitive cell like:
    ["Na", "Na", "Na", "Na", "Cl", "Cl", "Cl", "Cl1"] -> ["Na", "Cl"]

    """
    symbols = ph_nacl.unitcell.symbols
    symbols[-1] = "Cl1"
    masses = ph_nacl.unitcell.masses
    masses[-1] = 70.0
    cell = PhonopyAtoms(
        cell=ph_nacl.unitcell.cell,
        scaled_positions=ph_nacl.unitcell.scaled_positions,
        symbols=symbols,
        masses=masses,
    )
    with pytest.raises(RuntimeError) as e:
        get_primitive(cell, primitive_matrix="F")
    assert str(e.value).split("\n")[0] == "Atom symbol mapping failure."



"""QHA example of Si."""

import tarfile

import matplotlib.pyplot as plt
import numpy as np

from phonopy import Phonopy
from phonopy.interface.vasp import Vasprun, read_vasp


def get_force_sets(index):
    """Parse vasprun.xml and return forces."""
    with tarfile.open("vasprun_xmls.tar.lzma", "r:xz") as tr:
        with tr.extractfile("vasprun_xmls/vasprun.xml-%d" % index) as fp:
            vasprun = Vasprun(fp, use_expat=True)
            forces = vasprun.read_forces()
    return forces


def get_frequency(poscar_filename, force_sets):
    """Calculate phonons and return frequencies."""
    unitcell = read_vasp(poscar_filename)
    volume = unitcell.volume
    phonon = Phonopy(
        unitcell,
        [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
        primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
    )
    disps = np.zeros_like(force_sets)
    disps[0] = [0.01, 0, 0]
    phonon.dataset = {
        "natom": len(force_sets),
        "first_atoms": [
            {"number": 0, "displacement": [0.01, 0, 0], "forces": force_sets}
        ],
    }
    phonon.produce_force_constants()
    return phonon.get_frequencies([0.5, 0.5, 0]), volume


def main():
    """Run QHA."""
    frequencies = []
    volumes = []
    for i in range(-5, 6):
        poscar_filename = "POSCAR-%d" % i
        force_sets = get_force_sets(i)
        fs, v = get_frequency(poscar_filename, force_sets)
        frequencies.append(fs)
        volumes.append(v)

    for freq_at_X in np.array(frequencies).T:
        freq_squared = freq_at_X**2 * np.sign(freq_at_X)
        # np.sign is used to treat imaginary mode, since
        # imaginary frequency is returned as a negative value from phonopy.
        plt.plot(volumes, freq_squared, "o-")
        # for v, f2 in zip(volumes, freq_squared):
        #      print("%f %f" % (v, f2))
        # print('')
        # print('')
    plt.title("Frequeny squared (THz^2) at X-point vs volume (A^3)")
    plt.show()


if __name__ == "__main__":
    main()



"""Example of QHA calculation by Al."""

import numpy as np
import yaml
from yaml import CLoader as Loader

from phonopy import PhonopyQHA

volumes = []
energies = []
for line in open("e-v.dat"):
    v, e = line.split()
    volumes.append(float(v))
    energies.append(float(e))

entropy = []
cv = []
fe = []
for index in range(-5, 6):
    filename = "thermal_properties.yaml-%d" % index
    print("Reading %s" % filename)
    thermal_properties = yaml.load(open(filename), Loader=Loader)["thermal_properties"]
    temperatures = [v["temperature"] for v in thermal_properties]
    cv.append([v["heat_capacity"] for v in thermal_properties])
    entropy.append([v["entropy"] for v in thermal_properties])
    fe.append([v["free_energy"] for v in thermal_properties])

qha = PhonopyQHA(
    volumes,
    energies,
    temperatures=temperatures,
    free_energy=np.transpose(fe),
    cv=np.transpose(cv),
    entropy=np.transpose(entropy),
    t_max=400,
    verbose=True,
)

# qha.plot_helmholtz_volume().show()
# qha.plot_volume_temperature().show()
qha.plot_thermal_expansion().show()
# plot = qha.plot_volume_expansion()
# if plot:
#     plot.show()
# qha.plot_gibbs_temperature().show()
# qha.plot_bulk_modulus_temperature().show()
# qha.plot_heat_capacity_P_numerical().show()
# qha.plot_heat_capacity_P_polyfit().show()
# qha.plot_gruneisen_temperature().show()
